from graphene import ObjectType
from ..abstract_type_builder import AbstractSchema
from .types.queries.by_pk import ByPkQuery
from .types.queries.aggregate import AggregateQuery
from .types.queries.many import ManyQuery
from .types.mutations.delete import DeleteMutation
from .types.mutations.insert import InsertMutation
from .types.mutations.update import UpdateMutation
from .types.shared.where_clause import WhereClause
from .types.shared.order_clause import OrderClause
from .types.shared.where_predicate import WherePredicate
from .types.shared.basic import BasicType
from .types.shared.enum import EnumType
from .types.shared.pk_input import PkInputType
from .types.shared.aggregate_result import AggregateResultType
from .types.shared.aggregate_basic import AggregateBasicType
from .types.shared.aggregate_group_by import AggregateGroupByType
from .types.shared.insert_input_type import InsertInputType
from .types.shared.update_input_type import UpdateInputType
from .types.shared.list_update_input_type import ListUpdateInputType
from .types.shared.nested_item_update_input_type import (
    NestedItemUpdateInputType,
)
from .types.shared.nested_item_insert_input_type import (
    NestedItemInsertInputType,
)
from .types.shared.tag_to_pk import TagToPkType
from .resolvers.insert import InsertMutationResolverType
from .resolvers.update import UpdateMutationResolverType
from .resolvers.delete import DeleteMutationResolverType
from .mappings import Types, DJANGO_TYPE_NAMES_TO_GRAPHENE_TYPES


class DefaultSchema(AbstractSchema):

    DEFAULT_QUERIES = (ByPkQuery, ManyQuery, AggregateQuery)
    DEFAULT_MUTATIONS = (InsertMutation, UpdateMutation, DeleteMutation)

    BUILDER_MAP = {
        Types.AGGREGATE_RESULT_TYPE: AggregateResultType,
        Types.AGGREGATE_BASIC_TYPE: AggregateBasicType,
        Types.AGGREGATE_GROUP_BY_TYPE: AggregateGroupByType,
        Types.AGGREGATE_QUERY_TYPE: AggregateQuery,
        Types.BASIC_TYPE: BasicType,
        Types.BY_PK_QUERY_TYPE: ByPkQuery,
        Types.DELETE_MUTATION_RESOLVER: DeleteMutationResolverType,
        Types.DELETE_MUTATION_TYPE: DeleteMutation,
        Types.ENUM_TYPE: EnumType,
        Types.INSERT_INPUT_TYPE: InsertInputType,
        Types.INSERT_MUTATION_RESOLVER: InsertMutationResolverType,
        Types.INSERT_MUTATION_TYPE: InsertMutation,
        Types.LIST_UPDATE_INPUT_TYPE: ListUpdateInputType,
        Types.MANY_QUERY_TYPE: ManyQuery,
        Types.NESTED_ITEM_UPDATE_TYPE: NestedItemUpdateInputType,
        Types.NESTED_ITEM_INSERT_TYPE: NestedItemInsertInputType,
        Types.ORDER_CLAUSE_TYPE: OrderClause,
        Types.PK_INPUT_TYPE: PkInputType,
        Types.TAG_TO_PK_TYPE: TagToPkType,
        Types.UPDATE_INPUT_TYPE: UpdateInputType,
        Types.UPDATE_MUTATION_TYPE: UpdateMutation,
        Types.UPDATE_MUTATION_RESOLVER: UpdateMutationResolverType,
        Types.WHERE_CLAUSE_TYPE: WhereClause,
        Types.WHERE_PREDICATE_INPUT_TYPE: WherePredicate,
    }

    def __init__(self, field_type_map_overrides=None):
        super(DefaultSchema, self).__init__()
        self._field_type_map_overrides = field_type_map_overrides

    @property
    def FIELD_TYPE_MAP(self):
        if not hasattr(self, "_field_type_map"):
            self._field_type_map = DJANGO_TYPE_NAMES_TO_GRAPHENE_TYPES.copy()
            self._field_type_map.update(self._field_type_map_overrides or {})

        return self._field_type_map

    def get_node_name_for_model(self, model_class):
        return model_class.__name__

    def build_query_root_type(self, registry):
        return self._build_root_type(
            registry, "QueryRoot", "query_classes", self.DEFAULT_QUERIES
        )

    def build_mutation_root_type(self, registry):
        return self._build_root_type(
            registry,
            "MutationRoot",
            "mutation_classes",
            self.DEFAULT_MUTATIONS,
        )

    def _build_root_type(
        self, registry, type_name, root_type_classes_key, root_type_classes_default
    ):
        """Used to build a root type, e.g. Query or Mutation. Iterates through
        the Entity class's root-type field as specified by root_type_classes_key
        and checks whether the access_permissions field allows the root type to
        exist (i.e. we don't create an insertFoo mutation for a read-only FooEntity)
        and if so, updates the result with the fields returned by the get_root_fields
        method of the root-type.

        Args:
            type_name: str. One of "MutationRoot" or "QueryRoot"
            root_type_classes_key: str: One of "mutation_classes" or "query_classes"
            root_type_classes_default: A list root type classes, which will come from
                djraphql.schemas.default.types.queries|mutations.
        Returns:
            Graphene type representing the root type for GraphQL queries or mutations.
        """

        attrs = {}
        for model_class in registry.get_available_model_classes():
            entity_class = registry.get_entity_class(model_class)

            # Add the query/mutation fields to the root type being built.
            root_types = getattr(entity_class, root_type_classes_key)

            # If the Entity class does not override query_classes or mutation_classes,
            # then we use the DEFAULT_QUERIES or DEFAULT_MUTATIONS from the schema.
            if root_types is None:
                root_types = root_type_classes_default

            for root_type in root_types:
                # Check permissions before we add the query/mutation type to the root.
                # E.g. insertFoo requires CREATE permission, so we only want to
                # define the insertFoo operation if the FooEntity has a Create in
                # its access_permissions field.
                root_type_permissions = root_type.get_required_access_permissions()
                if not entity_class.allows_permissions(root_type_permissions):
                    continue

                attrs.update(root_type.get_root_fields(registry, model_class))

        return type(type_name, (ObjectType,), attrs)



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from graphene import List, InputObjectType, Boolean, Field
from ...mappings import Types
from graphql import Undefined


class WhereClause(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)

        # Build where predicate InputObjects to allow filtering in queries.
        where_input_type_attrs = {}
        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()

            if field.many_to_one or field.one_to_many:
                where_input_type_attrs[field.name] = Field(
                    registry.lambda_from_registry(
                        field.related_model, Types.WHERE_CLAUSE_TYPE
                    ),
                    default_value=Undefined,
                )

            if not field.concrete:
                continue

            # TODO: Remove when support for many to many is implemented
            if field.many_to_many:
                continue

            # Convert names of ForeignKey/OneToOneFields from 'foo' to 'foo_id'
            field_name = field.name
            if field.many_to_one or field.one_to_one and field_name != "id":
                field_name = "{}_id".format(field_name)

            # The predicate input types for e.g. IntegerFields need different type
            # definitions than CharFields.
            # e.g., where: {id: _eq: 1} vs. where: {name: _eq: "hello"}
            # Examine the type of a model's field and build the appropriately typed
            # predicate input objects.
            type_name = field.get_internal_type()
            if field.choices:
                enum_type = registry.get_or_create_type(
                    Types.ENUM_TYPE,
                    model_class=model_class,
                    field=field,
                    choices=field.choices,
                )
                where_input_type_attrs[field_name] = type(
                    "WherePredicate{}".format(enum_type.__name__),
                    (InputObjectType,),
                    {
                        "_eq": enum_type(name="_eq", default_value=Undefined),
                        "_neq": enum_type(name="_neq", default_value=Undefined),
                        "_in": List(enum_type, name="_in", default_value=Undefined),
                        "_nin": List(enum_type, name="_nin", default_value=Undefined),
                        "_is_null": Boolean(name="_is_null", default_value=Undefined),
                    },
                )()
            else:
                graphene_type = registry.get_graphene_type_or_default(model_field)
                where_input_type_attrs[field_name] = registry.get_or_create_type(
                    Types.WHERE_PREDICATE_INPUT_TYPE,
                    graphene_type=graphene_type,
                )(default_value=Undefined)

        # These fields are always the same, they just recursively reference the
        # input type we're building.
        where_input_lambda = registry.lambda_from_registry(
            model_class, Types.WHERE_CLAUSE_TYPE
        )
        where_input_type_attrs["_and"] = List(
            where_input_lambda, name="_and", default_value=Undefined
        )
        where_input_type_attrs["_or"] = List(
            where_input_lambda, name="_or", default_value=Undefined
        )
        where_input_type_attrs["_not"] = Field(
            where_input_lambda, name="_not", default_value=Undefined
        )

        return type(
            "Where{}".format(model_class.__name__),
            (InputObjectType,),
            where_input_type_attrs,
        )



from ..util.visitors import visit_query
from ..util.arguments import (
    DEFAULT_MAXIMUM_RESULTS,
    ABSOLUTE_MAXIMUM_RESULTS,
    update_queryset_based_on_arguments,
)


def build_list_resolver(model_class, registry, related_name=None):
    entity_class = registry.get_entity_class(model_class)

    def list_resolver(parent, info, **kwargs):

        # If related_name and parent are not None, then we are resolving a
        # subselect, which has already been prefetched via the code below.
        if related_name and parent is not None:
            # If we were given offset/limit parameters in the GraphQL query
            # we need to do them manually, because Django's ORM prevents
            # filtering _after_ slicing (rightfully, because it would break the
            # functionality of prefetching), which occurs naturally in
            # the course of query optimization: we'll make the select_related
            # and prefetch_related calls, but those prefetches work by
            # adding another filter predicate of the form "and x.id in (a, b, c)..."
            # which necessarily must occur after our optimization logic has run.
            # Thus, we must defer slicing until we enter this resolver.
            begin, end = get_begin_and_end_from_arguments(**kwargs)

            # Edge-case: if the request contained "twin queries", e.g.
            # LabelByPk(pk: 1) {
            #   oneArtist: artists(where: {id: {_eq: 1}}) { name }
            #   allArtists: artists(orderBy: {name: desc}) { name }
            # }
            # where we're querying the same field ("artists") multiple times,
            # it's handled as a "to_attr" during query optimization.
            # https://docs.djangoproject.com/en/3.1/ref/models/querysets/#prefetch-objects
            # This unfortunately materializes a list in memory _before_ slicing,
            # but it's a valuable use case so we need to handle it somehow.
            maybe_alias = info.path.key
            if maybe_alias != related_name and hasattr(parent, maybe_alias):
                return getattr(parent, maybe_alias)[begin:end]

            # There was a test for queryset.count() < 10000
            # here before. The idea was that if more than 10K rows were
            # about to be loaded into memory, it was better to N+1.
            # It was removed because it was untested and premature,
            # but possibly one day it may make sense to revisit.
            queryset = getattr(parent, related_name).get_queryset()
            return queryset[begin:end]

        # We're resolving a top-level select, so we must filter by org,
        # which is done by running the check lambda provided by the caller
        queryset = entity_class.get_queryset(info.context)
        artifacts = visit_query(registry, info)
        return artifacts["queryset_optimizer"].optimize(queryset)

    return list_resolver


def get_begin_and_end_from_arguments(**kwargs):
    begin = None
    end = None

    if "offset" in kwargs:
        begin = kwargs.get("offset") or 0

    if "limit" in kwargs:
        limit = min(kwargs["limit"], ABSOLUTE_MAXIMUM_RESULTS)
        end = (begin or 0) + limit

    return begin, end



from six import iteritems
from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from graphene import Field
from ...mappings import AGGREGATE_FIELD_NAMES_TO_DJANGO_FUNCTIONS, Types


class AggregateResultType(CacheableTypeBuilder, AbstractTypeBuilder):
    # @classmethod
    # def cache_key(cls, type_key, **kwargs):
    #     return type_key

    @staticmethod
    def make(registry, **kwargs):
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)
        attrs = {}
        for aggregate_field_name in AGGREGATE_FIELD_NAMES_TO_DJANGO_FUNCTIONS.keys():

            aggregate_basic_type = registry.get_or_create_type(
                Types.AGGREGATE_BASIC_TYPE,
                model_class=model_class,
                aggregate_field_name=aggregate_field_name,
            )

            attrs[aggregate_field_name] = Field(
                aggregate_basic_type, name=aggregate_field_name
            )
            attrs[
                "resolve_{}".format(aggregate_field_name)
            ] = resolver_for_aggregate_field(aggregate_field_name)

        group_by_type = registry.get_or_create_type(
            Types.AGGREGATE_GROUP_BY_TYPE, model_class=model_class
        )
        return type(
            "{}AggregateResult".format(model_class.__name__),
            (group_by_type,),
            attrs,
        )


class AggregateResultStruct:
    """Class is used for converting a dictionary to the equivalent object,
    which Graphene resolver expects. In aggregate (e.g. sum, max, etc.) queries,
    the results are dictionaries rather than model instances, so we must use this.
    """

    def __init__(self, **entries):
        for key, value in iteritems(entries):
            if isinstance(value, dict):
                self.__dict__[key] = AggregateResultStruct(**value)
            else:
                self.__dict__[key] = value


def resolver_for_aggregate_field(field_name):
    return lambda parent, info, **kwargs: AggregateResultStruct(**parent[field_name])



import logging
import traceback
from djraphql.django_models import get_pk_fields_for_model, get_field_by_name
from .by_pk import build_by_pk_resolver
from six import iteritems
from ..mappings import Types
from ...abstract_type_builder import CacheableTypeBuilder
from ..util.validation import perform_validation, FatalValidationError
from ..util import MutationAccumulator


class UpdateMutationResolverType(CacheableTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        return build_update_mutation_resolver(kwargs["model_class"], registry)


def build_update_mutation_resolver(model_class, registry):
    entity_class = registry.get_entity_class(model_class)

    def update_mutation_resolver(root, info, accumulator=None, **args):

        if not entity_class.is_updatable_by_context(info.context):
            raise Exception("Invalid permissions.")

        mutation_class = registry.get_or_create_type(
            Types.UPDATE_MUTATION_TYPE, model_class=model_class
        )

        # Perform validation
        is_valid, validation_errors = perform_validation(
            entity_class, args["data"], model_class, root, info
        )

        if validation_errors:
            accumulator.add_validation_error(validation_errors)

        if not is_valid:
            raise FatalValidationError()

        instance = _get_object_to_update(
            args, model_class, entity_class, info, registry
        )

        pre_rel_inputs, post_rel_inputs = _partition_relation_operations(
            args["data"], model_class, instance, info, registry
        )

        _handle_pre_save_relations(
            pre_rel_inputs, instance, accumulator, info, registry
        )

        # Set the updated values for our instance.
        for key, value in iteritems(args["data"]):
            # In Graphene 3, Enum fields are represented as an instance with
            # name and value properties. We want to save the value property.
            extracted_value = getattr(value, "value", value)
            setattr(instance, key, extracted_value)

        instance.save()

        if root is None:
            accumulator.root_result = instance

        _handle_post_save_relations(
            post_rel_inputs, instance, accumulator, info, registry
        )

        # TODO: this probably should be taken care of in the mutation class.
        resolved_result = instance
        if root is None:
            by_pk_resolver = build_by_pk_resolver(
                model_class,
                registry,
            )
            resolved_result = by_pk_resolver(None, info, pk=instance.pk)

        return mutation_class(
            success=True,
            pks_to_tags=accumulator.pks_to_tags,
            validation_errors=accumulator.validation_errors,
            result=resolved_result,
        )

    return update_mutation_resolver


def _handle_post_save_relations(post_rel_inputs, instance, accumulator, info, registry):
    # Iterate through each relational/nested input we found earlier
    # and call the resolver for each one. If the nested relationship is
    # a list, then call the resolver for each item in the list.
    for field, value in iteritems(post_rel_inputs):
        if field.one_to_many:
            perform_nested_list_update(
                registry,
                info,
                instance,
                value["items"],
                field,
                value["update_strategy"],
                accumulator,
            )
        elif field.one_to_one:
            updated_child = getattr(instance, field.name, None)
            if updated_child and value is None:
                if not registry.get_entity_class(field.related_model).is_deletable():
                    raise Exception(
                        "Setting {} to null is a DELETE operation, and {} is not deletable according to its schema permissions.".format(
                            field.name, field.related_model.__name__
                        )
                    )
                resolver = registry.get_or_create_type(
                    Types.DELETE_MUTATION_RESOLVER,
                    model_class=field.related_model,
                )
                resolver(instance, info, pk=updated_child.pk)
            elif updated_child:
                resolver = registry.get_or_create_type(
                    Types.UPDATE_MUTATION_RESOLVER, model_class=field.related_model
                )
                resolver(
                    instance,
                    info,
                    pk=updated_child.pk,
                    data=value,
                    accumulator=accumulator,
                )
            elif value is None:
                # Here we know updated_child is None, and since value is
                # also None, there's actually nothing to do, so just skip.
                continue
            else:
                value[field.field.name] = instance
                resolver = registry.get_or_create_type(
                    Types.INSERT_MUTATION_RESOLVER, model_class=field.related_model
                )
                resolver(instance, info, data=value, accumulator=accumulator)
        else:
            raise Exception("Not implemented")


def _handle_pre_save_relations(pre_rel_inputs, instance, accumulator, info, registry):
    # If the one-to-one relation object already exists, we call
    # update resolver, else call the insert resolver.
    for field, value in iteritems(pre_rel_inputs):
        relation_instance = getattr(instance, field.name, None)
        if relation_instance:
            resolver = registry.get_or_create_type(
                Types.UPDATE_MUTATION_RESOLVER,
                model_class=field.related_model,
            )
            resolver(
                instance,
                info,
                pk=relation_instance.pk,
                data=value,
                accumulator=accumulator,
            )
        else:
            resolver = registry.get_or_create_type(
                Types.INSERT_MUTATION_RESOLVER,
                model_class=field.related_model,
            )
            result = resolver(
                instance,
                info,
                data=value,
                accumulator=accumulator,
            )

            # Set the field on the instance to the inserted result
            setattr(instance, field.name, result.result)


def _partition_relation_operations(data, model_class, instance, info, registry):
    pre_save_rel_inputs = {}
    post_save_rel_inputs = {}

    data_keys = list(data.keys())
    for name in data_keys:
        field = get_field_by_name(model_class, name)
        if not field.is_relation:
            continue

        if not field.concrete:
            # Remove relational inputs from the dictionary passed to the model,
            # and handle them separately by calling a resolver for each one below.
            post_save_rel_inputs[field] = data.pop(field.name)
        else:
            # If the field is 1-1 and concrete, but is passed in as a dictionary,
            # we must save it before assigning it to the dict passed to the model.
            if isinstance(data[name], dict):
                value = data.pop(name)
                pre_save_rel_inputs[field] = value
            else:
                name_with_id = "{}_id".format(field.name)
                if name_with_id in data and field.name in data:
                    raise Exception(
                        'Pass one of "{}", "{}", not both.'.format(
                            name_with_id, field.name
                        )
                    )
                elif name_with_id in data and data[name_with_id] is not None:
                    # Check that the filter_backend actually includes the referenced object.
                    assert (
                        registry.get_entity_class(field.related_model)
                        .get_queryset(info.context)
                        .filter(pk=data[name_with_id])
                        .exists()
                    ), "{} matching query does not exist.".format(
                        field.related_model.__name__
                    )

    return pre_save_rel_inputs, post_save_rel_inputs


def _get_object_to_update(data, model_class, entity_class, info, registry):
    # Search for the updated item by queryset/PK to make sure it exists.
    assert "pk" in data, "Must pass pk for updates."
    pk_fields = registry.get_pk_fields_for_model(model_class)
    if len(pk_fields) == 1:
        pk_search_kwargs = {pk_fields[0].name: data["pk"]}
    else:
        pk_search_kwargs = {field.name: data[field.name] for field in pk_fields}

    # Fetch the instance by just its PK first, without any filtering from the
    # Entity class's filter_backend, which can result in multiple (duplicate)
    # objects being returned, whereas we always only want just one instance.
    # This was previously causing unnecessary errors during resolution, e.g.
    # ".get() returned more than 1, it returned 2!"
    instance = model_class.objects.get(**pk_search_kwargs)

    # Then, do the check against the Entity class's filter_backend.
    # This ensures we can only update/reference objects our filter_backend
    # has access to. Multiple (duplicate) objects can be returned by this query,
    # we only care that something is returned.
    assert (
        entity_class.get_queryset(info.context).filter(pk=instance.pk).exists()
    ), "{} matching query does not exist.".format(model_class.__name__)

    return instance


def perform_nested_list_update(
    registry,
    info,
    instance,
    items,
    field,
    update_strategy,
    accumulator,
):
    update_resolver = registry.get_or_create_type(
        Types.UPDATE_MUTATION_RESOLVER, model_class=field.related_model
    )

    insert_resolver = registry.get_or_create_type(
        Types.INSERT_MUTATION_RESOLVER, model_class=field.related_model
    )

    updated_or_inserted_pks = []
    accumulator.push_path_part(field.name)
    for i, item in enumerate(items):
        accumulator.push_path_part(i)
        if "pk" in item:
            updated_or_inserted_pks.append(item["pk"])
            update_resolver(instance, info, accumulator=accumulator, **item)
        else:
            item["data"][field.field.name] = instance
            result = insert_resolver(instance, info, accumulator=accumulator, **item)
            updated_or_inserted_pks.append(result.result.pk)
        accumulator.pop_path_part()
    accumulator.pop_path_part()

    # If our update strategy is anything other than 'replace', we're done.
    # Otherwise, we must delete any already-existing objects that do not
    # have a corresponding item in the mutation's nested list.
    if update_strategy != "replace":
        return

    items_to_delete = list(
        getattr(instance, field.name).exclude(pk__in=updated_or_inserted_pks)
    )

    # Check permissions and confirm that these objects are actually deletable
    if (
        items_to_delete
        and not registry.get_entity_class(field.related_model).is_deletable()
    ):
        raise Exception(
            "Nested update with UpdateStrategy 'replace' has resulted in a DELETE operation, but {} is not deletable according to its schema permissions.".format(
                field.name, field.related_model.__name__
            )
        )

    delete_resolver = registry.get_or_create_type(
        Types.DELETE_MUTATION_RESOLVER, model_class=field.related_model
    )

    for item in items_to_delete:
        delete_resolver(instance, info, pk=item.pk)



from django.db.models import Avg, Count, Max, Min, Sum
from graphene import String, ID, Int, Float, Boolean, DateTime, Date


DJANGO_TYPE_NAMES_TO_GRAPHENE_TYPES = {
    "AutoField": ID,
    "IntegerField": Int,
    "PositiveIntegerField": Int,
    "BigIntegerField": Int,
    "ForeignKey": ID,
    "OneToOneField": ID,
    "ManyToManyField": ID,
    "DecimalField": Float,
    "FloatField": Float,
    "DateField": Date,
    "DateTimeField": DateTime,
    "CharField": String,
    "TextField": String,
    "BooleanField": Boolean,
    "NullBooleanField": Boolean,
    "SmallIntegerField": Int,
}

AGGREGATE_FIELD_NAMES_TO_DJANGO_FUNCTIONS = {
    "_sum": Sum,
    "_min": Min,
    "_max": Max,
    "_avg": Avg,
    "_count": Count,
}


class Types:
    AGGREGATE_RESULT_TYPE = "aggregate_result_type"
    AGGREGATE_BASIC_TYPE = "aggregate_basic_type"
    AGGREGATE_GROUP_BY_TYPE = "aggregate_group_by_type"
    AGGREGATE_QUERY_TYPE = "aggregate_query_type"
    BASIC_TYPE = "basic_type"
    BY_PK_QUERY_TYPE = "by_pk_query_type"
    DELETE_MUTATION_RESOLVER = "delete_mutation_resolver"
    DELETE_MUTATION_TYPE = "delete_mutation_type"
    ENUM_TYPE = "enum_type"
    INSERT_INPUT_TYPE = "insert_input_type"
    INSERT_MUTATION_RESOLVER = "insert_mutation_resolver"
    INSERT_MUTATION_TYPE = "insert_mutation_type"
    LIST_UPDATE_INPUT_TYPE = "list_update_input_type"
    MANY_QUERY_TYPE = "many_query_type"
    NESTED_ITEM_INSERT_TYPE = "nested_item_insert_type"
    NESTED_ITEM_UPDATE_TYPE = "nested_item_update_type"
    ORDER_CLAUSE_TYPE = "order_clause_type"
    PK_INPUT_TYPE = "pk_input_type"
    UPDATE_INPUT_TYPE = "update_input_type"
    UPDATE_MUTATION_RESOLVER = "update_mutation_resolver"
    UPDATE_MUTATION_TYPE = "update_mutation_type"
    WHERE_CLAUSE_TYPE = "where_clause_type"
    WHERE_PREDICATE_INPUT_TYPE = "where_predicate_input_type"
    TAG_TO_PK_TYPE = "tag_to_pk_type"



from graphene import (
    List,
)
from djraphql.access_permissions import READ
from ....abstract_type_builder import QueryRootTypeBuilder, CacheableTypeBuilder
from ...resolvers.aggregate import build_aggregate_resolver
from ...util.arguments import create_query_arguments
from ...mappings import AGGREGATE_FIELD_NAMES_TO_DJANGO_FUNCTIONS, Types


class AggregateQuery(CacheableTypeBuilder, QueryRootTypeBuilder):
    @classmethod
    def make(cls, registry, **kwargs):
        model_class = kwargs["model_class"]
        registry.get_or_create_type(
            Types.AGGREGATE_GROUP_BY_TYPE, model_class=model_class
        )

        entity_class = registry.get_entity_class(model_class)

        where_type = registry.get_or_create_type(
            Types.WHERE_CLAUSE_TYPE, model_class=model_class
        )
        order_by_type = registry.get_or_create_type(
            Types.ORDER_CLAUSE_TYPE, model_class=model_class
        )
        field = List(
            registry.get_or_create_type(
                Types.AGGREGATE_RESULT_TYPE, model_class=model_class
            ),
            args=create_query_arguments(where_type, order_by_type),
            description=cls.get_docs(model_class),
        )

        resolver = build_aggregate_resolver(
            model_class, registry, AGGREGATE_FIELD_NAMES_TO_DJANGO_FUNCTIONS
        )

        return (
            field,
            resolver,
        )

    @staticmethod
    def get_field_name(model_class):
        model_name = model_class.__name__
        return "{}{}Aggregate".format(model_name, "" if model_name[-1] == "s" else "s")

    @staticmethod
    def get_docs(model_class):
        model_name = model_class.__name__
        return """Given (optional) filtering arguments (where, orderBy,
            limit, offset), returns a list of {} objects.
        """.format(
            model_name
        )

    @staticmethod
    def get_required_access_permissions():
        return READ



from graphql.language.visitor import visit, ParallelVisitor
from .queryset_optimization_visitor import QuerysetOptimizationVisitor
from .read_permissions_check_visitor import ReadPermissionsCheckVisitor


def visit_query(registry, info):
    query_visitor = QueryVisitor(
        [
            QuerysetOptimizationVisitor(registry, info),
            ReadPermissionsCheckVisitor(registry, info),
        ]
    )
    visit(info.field_nodes[0], query_visitor)
    return query_visitor.artifacts


class QueryVisitor(ParallelVisitor):
    @property
    def artifacts(self):
        result = {}
        for visitor in self.visitors:
            result.update(getattr(visitor, "artifacts", {}))
        return result



from graphene import (
    List,
    NonNull,
)
from djraphql.access_permissions import READ
from ....abstract_type_builder import QueryRootTypeBuilder, CacheableTypeBuilder
from ...resolvers.list import build_list_resolver
from ...mappings import Types
from ...util.arguments import create_query_arguments


class ManyQuery(CacheableTypeBuilder, QueryRootTypeBuilder):
    @classmethod
    def make(cls, registry, **kwargs):
        model_class = kwargs["model_class"]
        where_type = registry.get_or_create_type(
            Types.WHERE_CLAUSE_TYPE, model_class=model_class
        )
        order_by_type = registry.get_or_create_type(
            Types.ORDER_CLAUSE_TYPE, model_class=model_class
        )

        field = List(
            NonNull(registry.lambda_from_registry(model_class, Types.BASIC_TYPE)),
            args=create_query_arguments(where_type, order_by_type),
            description=cls.get_docs(model_class),
        )

        resolver = build_list_resolver(model_class, registry)

        return (field, resolver)

    @staticmethod
    def get_field_name(model_class):
        model_name = model_class.__name__
        return "{}{}Many".format(model_name, "" if model_name[-1] == "s" else "s")

    @staticmethod
    def get_docs(model_class):
        model_name = model_class.__name__
        return """Given (optional) filtering arguments (where, orderBy,
            limit, offset), returns a list of {} objects.
        """.format(
            model_name
        )

    @staticmethod
    def get_required_access_permissions():
        return READ



from abc import ABCMeta, abstractmethod, abstractproperty


class CacheableTypeBuilder:
    @classmethod
    def cache_key(cls, type_key, **kwargs):
        if "model_class" in kwargs:
            return "{}/{}".format(kwargs["model_class"].__name__, type_key)

        return type_key


class AbstractTypeBuilder:
    __metaclass__ = ABCMeta

    @staticmethod
    @abstractmethod
    def make(registry, **kwargs):
        raise NotImplementedError()


class AbstractRootTypeBuilder(AbstractTypeBuilder):
    @abstractmethod
    def get_field_name(self, model_class):
        pass

    @abstractmethod
    def get_docs(self, model_class):
        pass

    @abstractmethod
    def get_required_access_permissions():
        pass

    @classmethod
    @abstractmethod
    def get_root_fields(cls, registry, model):
        raise NotImplementedError()


class QueryRootTypeBuilder(AbstractRootTypeBuilder):
    @classmethod
    @abstractmethod
    def get_root_fields(cls, registry, model_class):
        field, resolver = cls.make(registry, model_class=model_class)
        field_name = cls.get_field_name(model_class)
        return {field_name: field, "resolve_{}".format(field_name): resolver}


class MutationRootTypeBuilder(AbstractRootTypeBuilder):
    @classmethod
    @abstractmethod
    def get_root_fields(cls, registry, model_class):
        mutation = cls.make(registry, model_class=model_class)
        mutation_name = cls.get_field_name(model_class)
        return {mutation_name: mutation.Field()}


class AbstractSchema:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_node_name_for_model(self, model_class):
        raise NotImplementedError()

    @abstractmethod
    def build_query_root_type(self, registry):
        raise NotImplementedError()

    @abstractmethod
    def build_mutation_root_type(self, registry):
        raise NotImplementedError()

    @abstractproperty
    def BUILDER_MAP(self):
        raise NotImplementedError()

    @abstractproperty
    def FIELD_TYPE_MAP(self):
        raise NotImplementedError()

    @abstractproperty
    def DEFAULT_QUERIES(self):
        raise NotImplementedError()

    @abstractproperty
    def DEFAULT_MUTATIONS(self):
        raise NotImplementedError()





# def test_fetched_fields_visitor_gets_scalar_fields():
#     ast = parse("""
#         query {
#             LabelsMany(where: {name: {_eq: "Parlophone"}}) {
#                 name
#                 establishedYear
#                 artists {
#                     name
#                     albums(where: { title: { _eq: "Help!" } }) {
#                         title
#                     }
#                 }
#             }
#         }
#     """)
#     visitor = FetchedFieldsVisitor()
#     result = visit(ast, visitor)
#     assert visitor.fields_by_key is None



from graphene import (
    Field,
    Argument,
)
from ....abstract_type_builder import QueryRootTypeBuilder, CacheableTypeBuilder
from ...resolvers.by_pk import build_by_pk_resolver
from ...mappings import Types
from djraphql.access_permissions import READ


class ByPkQuery(CacheableTypeBuilder, QueryRootTypeBuilder):
    @classmethod
    def make(cls, registry, **kwargs):
        model_class = kwargs["model_class"]
        basic_type = registry.get_or_create_type(
            Types.BASIC_TYPE, model_class=model_class
        )
        pk_input_type = registry.get_or_create_type(
            Types.PK_INPUT_TYPE, model_class=model_class
        )
        field = Field(
            basic_type,
            args={"pk": Argument(pk_input_type, required=True)},
            description=cls.get_docs(model_class),
        )

        resolver = build_by_pk_resolver(model_class, registry)

        return (
            field,
            resolver,
        )

    @staticmethod
    def get_field_name(model_class):
        model_name = model_class.__name__
        return "{}ByPk".format(model_name)

    @staticmethod
    def get_docs(model_class):
        model_name = model_class.__name__
        return "Given its primary-key, returns a single {} object.".format(model_name)

    @staticmethod
    def get_required_access_permissions():
        return READ



from graphene import Int, String
from djraphql.django_models import get_field_by_name


class ModelField:
    def __init__(self, name, read_only=False, graphene_type=None):
        self.name = name
        self.read_only = read_only
        self.graphene_type = graphene_type

    def get_resolver(self, entity_class):
        return lambda context, instance: getattr(instance, self.name, None)

    def get_django_model_field(self):
        return get_field_by_name(self.model_class, self.name)

    def with_model_class(self, model_class):
        """This method is used to hydrate this ModelField instance
        with the associated model class (which comes from Entity.meta.model).

        Maybe a bit hacky, but if we didn't have it, we'd have to provide the
        model class to each instance in the Entity.Meta.fields tuple,
        which is a poor DX.
        """

        self.model_class = model_class
        return self


class ComputedField:
    def __init__(self, name, graphene_type, depends_on=()):
        self.name = name
        self.graphene_type = graphene_type
        self.depends_on = depends_on

    def get_graphene_type(self):
        return self.graphene_type

    def get_resolver(self, entity_class):
        def method_property_resolver(context, instance):
            resolver = getattr(entity_class, "get_{}".format(self.name))
            return resolver(context, instance)

        return method_property_resolver


class AllModelFields:
    def __init__(self, excluding=()):
        self.exclude_fields = excluding

    def field_is_excluded(self, field):
        return field.name in self.exclude_fields

    def generate_model_fields(self, model_class):
        return {
            field.name: ModelField(field.name).with_model_class(model_class)
            for field in model_class._meta.get_fields()
            if field.name not in self.exclude_fields
        }



from djraphql.django_models import get_pk_fields_for_model
from ..mappings import Types
from ...abstract_type_builder import CacheableTypeBuilder


class DeleteMutationResolverType(CacheableTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        return build_delete_mutation_resolver(kwargs["model_class"], registry)


def build_delete_mutation_resolver(model_class, registry):
    entity_class = registry.get_entity_class(model_class)

    def delete_mutation_resolver(root, info, **args):

        if not entity_class.is_deletable_by_context(info.context):
            raise Exception("Invalid permissions.")

        # Search for the deleted item by queryset/PK to make sure it exists.
        assert "pk" in args, "Must pass pk for deletes."
        pk_fields = registry.get_pk_fields_for_model(model_class)

        if len(pk_fields) == 1:
            pk_search_kwargs = {pk_fields[0].name: args["pk"]}
        else:
            pk_search_kwargs = {field.name: args[field.name] for field in pk_fields}

        queryset = entity_class.get_queryset(info.context)
        instance = queryset.get(**pk_search_kwargs)

        success = False
        try:
            instance.delete()
            success = True
        except Exception as e:
            print("Something bad happened!", e)

        mutation_class = registry.get_or_create_type(
            Types.DELETE_MUTATION_TYPE, model_class=model_class
        )
        return mutation_class(success=success)

    return delete_mutation_resolver



from djraphql.django_models import get_target_field
from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from ...mappings import Types
from graphene import Field, InputObjectType, List
from graphql import Undefined


class InsertInputType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        """Generates a Graphene InputObjectType for the associated model.
        https://docs.graphene-python.org/en/latest/types/mutations/#inputfields-and-inputobjecttypes
        """
        input_class_attrs = {}
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)

        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()

            if (field.concrete and field.primary_key) or model_field.read_only:
                continue

            if field.one_to_many:
                related_entity = registry.get_entity_class(field.related_model)
                if related_entity.is_creatable():
                    input_class_attrs[field.name] = List(
                        registry.get_or_create_type(
                            Types.NESTED_ITEM_INSERT_TYPE,
                            model_class=field.related_model,
                        ),
                        default_value=Undefined,
                    )
            elif field.many_to_one or field.one_to_one:
                related_entity = registry.get_entity_class(field.related_model)
                if related_entity.can_get_for_insert():
                    continue

                # If the field is concrete, define a *_id field so that, e.g., we
                # can insert an Album with a labelId.
                if field.concrete:
                    reverse_model_field = registry.get_model_fields(
                        field.related_model
                    )[get_target_field(field).name]

                    graphene_type = registry.get_graphene_type_or_default(
                        reverse_model_field
                    )
                    input_class_attrs["{}_id".format(field.name)] = graphene_type(
                        default_value=Undefined
                    )

                # Check that the related object type is creatable and if it is, allow
                # a nested create by defining the related model's insert input type.
                if field.is_relation and related_entity.is_creatable():
                    input_class_attrs[field.name] = Field(
                        registry.lambda_from_registry(
                            field.related_model, Types.INSERT_INPUT_TYPE
                        ),
                        default_value=Undefined,
                    )
            elif field.many_to_many:
                pass
            elif field.choices:
                enum_type = registry.get_or_create_type(
                    Types.ENUM_TYPE,
                    model_class=model_class,
                    field=field,
                    choices=field.choices,
                )
                input_class_attrs[field.name] = Field(
                    enum_type,
                    required=not field.null and field.default is None,
                    default_value=Undefined,
                )
            else:
                input_class_attrs[field.name] = registry.get_graphene_type_or_default(
                    model_field
                )(
                    required=not field.null and not field.has_default(),
                    default_value=Undefined,
                )

        return type(
            "{}Input".format(model_class.__name__),
            (InputObjectType,),
            input_class_attrs,
        )



from djraphql.django_models import get_target_field
from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from ...mappings import Types
from graphene import Field, InputObjectType
from graphql import Undefined


class UpdateInputType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        """Generates a Graphene InputObjectType for the associated model.
        https://docs.graphene-python.org/en/latest/types/mutations/#inputfields-and-inputobjecttypes
        """
        input_class_attrs = {}
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)
        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()

            # Don't allow updating the field is it is (part of) a PK,
            # or it's not included in the Entity.Meta.fields list,
            # or it is read-only.
            if (field.concrete and field.primary_key) or model_field.read_only:
                continue

            if (
                field.is_relation
                and not registry.get_entity_class(field.related_model).is_updatable()
            ):
                # We don't allow the related object to be updated, but we do allow
                # the reference to it to change, i.e., if a Foo has a FK to a Bar,
                # we allow foo.bar_id to be updated (given that a Foo is updatable).
                if field.concrete and not field.many_to_many:
                    # If the related model's entity overrides get_for_insert,
                    # that implies that this field is read-only.
                    related_entity = registry.get_entity_class(field.related_model)
                    if related_entity.can_get_for_insert():
                        continue

                    reverse_model_field = registry.get_model_fields(
                        field.related_model
                    )[get_target_field(field).name]

                    graphene_type = registry.get_graphene_type_or_default(
                        reverse_model_field
                    )
                    input_class_attrs["{}_id".format(field.name)] = graphene_type(
                        default_value=Undefined
                    )
                continue

            if field.one_to_many:
                list_update_input_type = registry.get_or_create_type(
                    Types.LIST_UPDATE_INPUT_TYPE, model_class=field.related_model
                )
                input_class_attrs[field.name] = Field(
                    list_update_input_type, default_value=Undefined
                )
            elif field.many_to_one or field.one_to_one:
                if field.concrete:
                    reverse_model_field = registry.get_model_fields(
                        field.related_model
                    )[get_target_field(field).name]

                    graphene_type = registry.get_graphene_type_or_default(
                        reverse_model_field
                    )
                    input_class_attrs["{}_id".format(field.name)] = graphene_type(
                        default_value=Undefined
                    )
                related_model = field.related_model
                input_class_attrs[field.name] = Field(
                    registry.lambda_from_registry(
                        related_model, Types.UPDATE_INPUT_TYPE
                    ),
                    default_value=Undefined,
                )
            elif field.many_to_many:
                pass
            elif field.choices:
                enum_type = registry.get_or_create_type(
                    Types.ENUM_TYPE,
                    model_class=model_class,
                    field=field,
                    choices=field.choices,
                )
                input_class_attrs[field.name] = Field(
                    enum_type, required=False, default_value=Undefined
                )
            else:
                input_class_attrs[field.name] = registry.get_graphene_type_or_default(
                    model_field
                )(default_value=Undefined)

        return type(
            "{}UpdateInput".format(model_class.__name__),
            (InputObjectType,),
            input_class_attrs,
        )



from .type_registry import TypeRegistry
from .schemas.default import DefaultSchema


class SchemaBuilder(object):
    """Entry point into the library. It is responsible for passing Entity classes
    to the TypeRegistry instance (via register_entity_classes) and building the
    root Query and Mutation types.

    Exposes QueryRoot and MutationRoot, which should then be passed to a
    graphene.Schema instance.
    """

    def __init__(self, schema=None):
        self.schema = schema or DefaultSchema()
        self.registry = TypeRegistry(self.schema)

    def register_entity_classes(self, *entity_classes):
        self.registry.register_entity_classes(*entity_classes)

    @property
    def QueryRoot(self):
        return self.schema.build_query_root_type(self.registry)

    @property
    def MutationRoot(self):
        return self.schema.build_mutation_root_type(self.registry)



import logging
import traceback
from djraphql.django_models import get_field_by_name
from six import iteritems
from ..mappings import Types
from ...abstract_type_builder import CacheableTypeBuilder
from .update import perform_nested_list_update
from ..util import MutationAccumulator
from ..util.validation import perform_validation, FatalValidationError


class InsertMutationResolverType(CacheableTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        return build_insert_mutation_resolver(kwargs["model_class"], registry)


def build_insert_mutation_resolver(model_class, registry):
    entity_class = registry.get_entity_class(model_class)

    def insert_mutation_resolver(root, info, accumulator=None, **args):

        if not entity_class.is_creatable_by_context(info.context):
            raise Exception("Invalid permissions.")

        mutation_class = registry.get_or_create_type(
            Types.INSERT_MUTATION_TYPE, model_class=model_class
        )

        # Perform validation
        is_valid, validation_errors = perform_validation(
            entity_class, args["data"], model_class, root, info, is_insert=True
        )
        if validation_errors:
            accumulator.add_validation_error(validation_errors)

        if not is_valid:
            raise FatalValidationError()

        pre_save_rel_inputs, post_save_rel_inputs = _partition_relation_operations(
            args["data"], info.context, model_class, registry
        )

        args["data"].update(
            _call_any_get_for_inserts(model_class, info.context, registry)
        )

        entity_class.before_insert(args["data"], info.context)

        # In Graphene 3, Enum fields are represented as an instance with
        # name and value properties. We want to save the value property.
        instance = model_class(
            **{k: getattr(v, "value", v) for k, v in iteritems(args["data"])}
        )

        _handle_pre_save_relations(
            instance,
            pre_save_rel_inputs,
            registry,
            info,
            accumulator,
        )

        instance.save()

        # If the query passed a tag attribute, map it to the inserted object's PK.
        # But only on the root object; the nested list inserts are handled
        # via perform_nested_list_update below.
        if args.get("tag"):
            accumulator.add_pk_to_tag(
                instance.pk, args["tag"], instance.__class__.__name__
            )

        if root is None:
            accumulator.root_result = instance

        _handle_post_save_relations(
            instance,
            post_save_rel_inputs,
            registry,
            info,
            accumulator,
        )

        entity_class.after_insert(instance, info.context)

        return mutation_class(
            success=True,
            pks_to_tags=accumulator.pks_to_tags,
            validation_errors=accumulator.validation_errors,
            result=instance,
        )

    return insert_mutation_resolver


def _partition_relation_operations(data, context, model_class, registry):
    pre_save_rel_inputs = {}
    post_save_rel_inputs = {}
    data_keys = list(data.keys())
    for name in data_keys:
        field = get_field_by_name(model_class, name)
        if not field.is_relation:
            continue

        if field.concrete:
            name_with_id = "{}_id".format(field.name)
            if name_with_id in data and field.name in data:
                raise Exception(
                    'Pass one of "{}", "{}", not both.'.format(name_with_id, field.name)
                )
            elif name_with_id in data:
                # If the *_id is None, skip the following check, since None doesn't exist
                if data[name_with_id] is None:
                    continue

                # Otherwise, check that the filter_backend actually includes the referenced object
                related_entity_class = registry.get_entity_class(field.related_model)
                assert (
                    related_entity_class.get_queryset(context)
                    .filter(pk=data[name_with_id])
                    .exists()
                ), "{} matching query does not exist.".format(
                    field.related_model.__name__
                )
            else:
                pre_save_rel_inputs[field] = data.pop(field.name)
        else:
            post_save_rel_inputs[field] = data.pop(field.name)

    return pre_save_rel_inputs, post_save_rel_inputs


def _call_any_get_for_inserts(model_class, context, registry):
    result = {}
    # Update the data dict based on any caller-provided relation values
    for field in model_class._meta.get_fields():
        if not field.concrete or not field.is_relation:
            continue

        related_entity_class = registry.get_entity_class(field.related_model)
        if (
            related_entity_class is None
            or not related_entity_class.can_get_for_insert()
        ):
            continue

        result[field.name] = related_entity_class.get_for_insert(context)

    return result


def _handle_pre_save_relations(
    instance, pre_save_rel_inputs, registry, info, accumulator
):
    # Save the relations whose PK instance needs to be valid
    for field, value in iteritems(pre_save_rel_inputs):
        resolver = registry.get_or_create_type(
            Types.INSERT_MUTATION_RESOLVER, model_class=field.related_model
        )

        # It is possible that field is e.g. a concrete OneToOneField,
        # which means we must save the model before calling setattr,
        # because value will be a dict instead of a model.
        if isinstance(value, dict):
            accumulator.push_path_part(field.name)
            result = resolver(
                None,
                info,
                data=value,
                accumulator=accumulator,
            )
            accumulator.pop_path_part()
            setattr(instance, field.name, result.result)
        else:
            setattr(instance, field.name, value)


def _handle_post_save_relations(
    instance, post_save_rel_inputs, registry, info, accumulator
):
    # Save relations that need the instance's PK to be valid
    for field, value in iteritems(post_save_rel_inputs):
        resolver = registry.get_or_create_type(
            Types.INSERT_MUTATION_RESOLVER, model_class=field.related_model
        )
        if field.one_to_many:
            # value is a list when this is an insert* mutation,
            # and a dict when it's an update* mutation.
            # This is because for update* mutations, we allow the user
            # to specify an update stategy (merge or replace) which drives
            # whether each item in the list is created/updated/deleted.
            # But for an insert, all we need is the list, because we know
            # there are no nested items to update/delete.
            items = value
            update_strategy = "merge"
            if isinstance(value, dict):
                items = value["items"]
                update_strategy = value["update_strategy"]
            perform_nested_list_update(
                registry,
                info,
                instance,
                items,
                field,
                update_strategy,
                accumulator,
            )
        else:
            # It is possible that field is e.g. a concrete OneToOneField,
            # which means we must save the model before calling setattr,
            # because value will be a dict instead of a model.
            if isinstance(value, dict):
                value[field.field.name] = instance
                accumulator.push_path_part(field.name)
                result = resolver(
                    None,
                    info,
                    data=value,
                    accumulator=accumulator,
                )
                accumulator.pop_path_part()
                setattr(instance, field.name, result.result)
            else:
                setattr(instance, field.name, value)



from djraphql.access_permissions import (
    CREATE,
    READ,
    UPDATE,
    DELETE,
    R,
    AccessPermissionsBitmap,
)
from djraphql.fields import AllModelFields, ComputedField, ModelField


class Entity:
    properties = ()
    filter_backends = ()
    access_permissions = (R,)
    query_classes = None
    mutation_classes = None
    validator = None

    @classmethod
    def get_queryset(cls, context):
        queryset = cls.Meta.model.objects.all()

        try:
            for backend_filter in cls.filter_backends:
                queryset = backend_filter.filter_backend(context, queryset)
        except Exception as e:
            raise Exception(
                "An error occurred when applying filter backend {}: {}".format(
                    backend_filter.__class__.__name__, str(e)
                )
            )

        return queryset

    @classmethod
    def can_get_for_insert(cls):
        return cls.get_for_insert != Entity.get_for_insert

    @staticmethod
    def get_for_insert(context):
        raise NotImplementedError()

    @classmethod
    def allows_permissions(cls, access_permissions):
        """Check that the access_permissions field defined on this Entity class
        matches the passed in access_permissions bitmap, which will come from
        AbstractRootTypeBuilder.get_required_access_permissions().

        This check is generally used to ensure we want to expose a Root query
        or mutation for an Entity, e.g. fooByPk should only exist if FooEntity's
        READ flag is 1.
        """
        return access_permissions == (
            cls._get_access_permissions_bitmap() & access_permissions
        )

    @classmethod
    def is_creatable(cls):
        return bool(cls._get_access_permissions_bitmap() & CREATE)

    @classmethod
    def is_readable(cls):
        return bool(cls._get_access_permissions_bitmap() & READ)

    @classmethod
    def is_updatable(cls):
        return bool(cls._get_access_permissions_bitmap() & UPDATE)

    @classmethod
    def is_deletable(cls):
        return bool(cls._get_access_permissions_bitmap() & DELETE)

    @classmethod
    def is_creatable_by_context(cls, context):
        return cls._is_context_authorized_for_permission_type(context, CREATE)

    @classmethod
    def is_readable_by_context(cls, context):
        return cls._is_context_authorized_for_permission_type(context, READ)

    @classmethod
    def is_updatable_by_context(cls, context):
        return cls._is_context_authorized_for_permission_type(context, UPDATE)

    @classmethod
    def is_deletable_by_context(cls, context):
        return cls._is_context_authorized_for_permission_type(context, DELETE)

    @classmethod
    def _is_context_authorized_for_permission_type(cls, context, permission_type):
        for access_permission in cls.access_permissions:
            if access_permission.matches(permission_type):
                return access_permission.has_permission(context)

        return False

    @classmethod
    def _get_access_permissions_bitmap(cls):
        # Cache the AccessPermissionsBitmap and reuse it across calls
        if not hasattr(cls, "_access_permissions_bitmap"):
            cls._access_permissions_bitmap = AccessPermissionsBitmap(
                *cls.access_permissions
            )
        return cls._access_permissions_bitmap

    @staticmethod
    def before_insert(data, context):
        pass

    @staticmethod
    def after_insert(instance, context):
        pass

    @classmethod
    def _get_meta_fields(cls):
        return getattr(cls.Meta, "fields", ())

    @classmethod
    def _get_computed_fields(cls):
        return [f for f in cls._get_meta_fields() if isinstance(f, ComputedField)]

    @classmethod
    def get_all_model_fields(cls):
        result = {}
        model_fields, _, special_fields = cls._partition_basic_and_special_fields()

        # Get ModelFields that are implicit via AllModelFields
        if special_fields:
            implicit_fields = special_fields[0].generate_model_fields(cls.Meta.model)
            result.update(implicit_fields)

        # Get the raw ModelFields define by the user, which override implicit ones
        for explicit_field in model_fields:
            if special_fields and explicit_field.name not in result:
                raise Exception(
                    "{}.{} (a) does not exist, or (b) both excluded via AllModelFields's excluding keyword argument and included as a ModelField, or (c) references a model that is not specified by an Entity.".format(
                        cls.Meta.model.__name__, explicit_field.name
                    )
                )
            result[explicit_field.name] = explicit_field.with_model_class(
                cls.Meta.model
            )

        return result.values()

    @classmethod
    def _partition_basic_and_special_fields(cls):
        fields_list = cls._get_meta_fields()
        special_fields = []
        computed_fields = []
        basic_fields = []

        for field in fields_list:
            if isinstance(field, AllModelFields):
                special_fields.append(field)
            elif isinstance(field, ComputedField):
                computed_fields.append(field)
            else:
                basic_fields.append(field)

        assert len(special_fields) <= 1, "Provide zero or one AllModelFields instances."
        return basic_fields, computed_fields, special_fields



from graphene import Int, List, Argument, Boolean
from graphql.pyutils import Undefined
from six import iteritems
from django.conf import settings
from django.db.models import Q
from django.db.models.functions import Lower
from functools import reduce

SCHEMA_BUILDER_SETTINGS = getattr(settings, "SCHEMA_BUILDER", {})
# default number of records retrieved for list query if no limit specified
DEFAULT_MAXIMUM_RESULTS = SCHEMA_BUILDER_SETTINGS.get("DEFAULT_MAXIMUM_RESULTS", 50)
# maximum number of records a list query can retrieve
ABSOLUTE_MAXIMUM_RESULTS = SCHEMA_BUILDER_SETTINGS.get("ABSOLUTE_MAXIMUM_RESULTS", 500)


def create_query_arguments(where_type, order_by_type):
    return {
        "where": Argument(where_type, default_value=Undefined),
        "order_by": Argument(List(order_by_type), default_value=Undefined),
        "limit": Argument(Int, default_value=Undefined),
        "offset": Argument(Int, default_value=Undefined),
        "distinct": Argument(Boolean, default_value=Undefined),
    }


def update_queryset_based_on_arguments(queryset, args, take_slice=True):
    if "where" in args:
        annotations = dict()
        q = filter_from_where_args(args["where"], annotations)
        queryset = queryset.annotate(**annotations).filter(q)

    if "order_by" in args:
        sort_order = []
        for sort_entry in args["order_by"]:
            sort_keys = list(sort_entry.keys())
            assert len(sort_keys) == 1, "order_by items can only have 1 key."
            sort_key = sort_keys[0]
            sort_values = list(sort_entry.values())
            sort_direction = sort_values[0]
            dash = "" if sort_direction.value == "ascending" else "-"
            sort_order.append("{}{}".format(dash, sort_key))
        queryset = queryset.order_by(*sort_order)

    if args.get("distinct"):
        queryset = queryset.distinct()

    if take_slice:
        limit = min(
            args.get("limit", DEFAULT_MAXIMUM_RESULTS), ABSOLUTE_MAXIMUM_RESULTS
        )
        offset = args.get("offset", 0)
        queryset = queryset[offset : offset + limit]

    return queryset


def filter_from_where_args(where_args, annotations, _and=True):
    combined_qs = []
    for key, value in iteritems(where_args):
        # Conjunction operations
        if key in ("_and", "_or"):
            joined = reduce(
                lambda result, q: q & result if key == "_and" else q | result,
                [filter_from_where_args(item, annotations) for item in value],
            )
            combined_qs.append(joined)

        # Negation operation
        if key == "_not":
            combined_qs.append(~filter_from_where_args(value, annotations))

        # Predicate operations
        if "_eq" in value:
            combined_qs.append(Q(**{key: get_raw_value(value["_eq"])}))
        if "_neq" in value:
            combined_qs.append(~Q(**{key: get_raw_value(value["_neq"])}))
        if "_gt" in value:
            combined_qs.append(Q(**{"{}__gt".format(key): get_raw_value(value["_gt"])}))
        if "_gte" in value:
            combined_qs.append(
                Q(**{"{}__gte".format(key): get_raw_value(value["_gte"])})
            )
        if "_lt" in value:
            combined_qs.append(Q(**{"{}__lt".format(key): get_raw_value(value["_lt"])}))
        if "_lte" in value:
            combined_qs.append(
                Q(**{"{}__lte".format(key): get_raw_value(value["_lte"])})
            )
        if "_in" in value:
            combined_qs.append(Q(**{"{}__in".format(key): get_raw_value(value["_in"])}))
        if "_nin" in value:
            combined_qs.append(
                ~Q(**{"{}__in".format(key): get_raw_value(value["_nin"])})
            )
        if "_is_null" in value:
            combined_qs.append(
                Q(**{"{}__isnull".format(key): get_raw_value(value["_is_null"])})
            )
        if "_like" in value:
            combined_qs.append(
                Q(**{"{}__contains".format(key): get_raw_value(value["_like"])})
            )
        if "_ilike" in value:
            annotations["{}_lower".format(key)] = Lower(key)
            combined_qs.append(
                Q(
                    **{
                        "{}_lower__icontains".format(key): get_raw_value(
                            value["_ilike"]
                        ).lower()
                    }
                )
            )

    if not combined_qs:
        # If we're here, the where clause contains a nested join, e.g.
        # albums(where: { artist: { label: { name: {_eq: "Record Co."}}}})
        # which means we need to generate Q(artist__label__name='Record Co.')
        for next_key, next_value in iteritems(value):
            nested_key = "{}__{}".format(key, next_key)
            combined_qs.append(
                filter_from_where_args({nested_key: next_value}, annotations)
            )

    if not combined_qs:
        print("Warning: combined_qs should never be empty at this point.")

    # Join them all together using ANDs or ORs.
    return reduce(lambda result, q: q & result if _and else q | result, combined_qs)


def get_raw_value(value):
    if isinstance(value, list):
        return [get_raw_value(v) for v in value]

    if hasattr(value, "value"):
        return value.value

    return value



from django.db.models.query import Prefetch
from .arguments import update_queryset_based_on_arguments


class QuerysetOptimizer:
    def __init__(self, root_info, selection_prefetch_info):
        """One of the artifacts produced by the visit_query method in the
        visitors/__init__.py file is an instance of this class.

        Args:
            root_info: NodeInfo instance. The NodeInfo class is defined in the
                visitors/query_optimization_visitor.py file.
            selection_prefecth_info: dict where the keys are Django model
                paths to be used as arguments into Prefetch instances, and the
                values are lists of NodeInfo instances.
        """
        self.root_info = root_info
        self.selection_prefetch_info = selection_prefetch_info

    def optimize(self, queryset):
        """Given the info passed to our constructor, take the given queryset
        and apply the select_related and prefetch_related calls so that we can
        minimize the number of SQL queries we need to fulfill the request.

        Additionally, call .only() for each one to ensure we only fetch the
        columns necessary to fulfill the request.
        """

        # Why do we sort the keys? Because we ran into an issue where
        # if we process e.g. ['artists__albums', 'artists'], Django complains:
        # ValueError("'artists' lookup was already seen with a different queryset.
        # You may need to adjust the ordering of your lookups.",)
        # Oddly, the error does not happen when processed in order
        # like ['artists', 'artists__albums']. So we sort the keys. This will
        # probably blow up in unexpected ways but theoretically the order shouldn't
        # matter at all, so if ordering them to appease Django works, it feels OK.
        to_many_sorted_keys = sorted(self.selection_prefetch_info.keys())
        for key in to_many_sorted_keys:
            prefetches = []
            use_to_attr = len(self.selection_prefetch_info[key]) > 1
            for prefetch_info in self.selection_prefetch_info[key]:

                # Only select the columns necessary to fulfill the request
                prefetch_queryset = prefetch_info.model_class.objects.only(
                    *prefetch_info.field_names
                )

                # Prefetches (which generally map to *-to-many relations of the root
                # object) can have nested *-to-one relations which we need to fetch
                # via select_related. This minimizes the # of queries Django ORM executes.
                if prefetch_info.select_related_keys:
                    prefetch_queryset = prefetch_queryset.select_related(
                        *prefetch_info.select_related_keys
                    )

                prefetches.append(
                    Prefetch(
                        prefetch_info.alias_path if use_to_attr else key,
                        queryset=update_queryset_based_on_arguments(
                            prefetch_queryset,
                            prefetch_info.arguments,
                            take_slice=False,
                        ),
                        to_attr=prefetch_info.alias if use_to_attr else None,
                    )
                )

            # Call prefetch_related, passing in the Prefetch object that contains
            # our optimizations plus filtering for any arguments from the query.
            queryset = queryset.prefetch_related(*prefetches)

        # The prefetch_stack should now have a single item, which maps
        # to the root object in the query. Select only necessary columns and filter
        # via any arguments passed in.
        queryset = update_queryset_based_on_arguments(
            queryset.only(*self.root_info.field_names),
            self.root_info.arguments,
        )

        # Finally, add any necessary select_related calls.
        if self.root_info.select_related_keys:
            queryset = queryset.select_related(*self.root_info.select_related_keys)

        return queryset



import re
from six import string_types
from ....abstract_type_builder import AbstractTypeBuilder
from graphene import Enum


class EnumType(AbstractTypeBuilder):
    @classmethod
    def cache_key(cls, type_key, **kwargs):
        model_class = kwargs["model_class"]
        field = kwargs["field"]
        return "enum/{model_class}/{field}".format(
            model_class=model_class.__name__, field=field.name
        )

    @staticmethod
    def make(registry, **kwargs):
        model_class = kwargs["model_class"]
        field = kwargs["field"]
        choices = kwargs["choices"]
        entity_class = registry.get_entity_class(model_class)
        # Values can be strings or integers. We basically always want the attribute
        # keys to be a descriptive string. So we don't like to use integers
        enum_attrs = {}
        for value, pretty in choices:
            name = value if isinstance(value, string_types) else pretty
            enum_name = format_graphql_enum_name(name)
            enum_attrs[enum_name] = value

        camel_case = "".join([part.capitalize() for part in field.name.split("_")])
        return type(
            "{}{}Enum".format(model_class.__name__, camel_case),
            (Enum,),
            enum_attrs,
        )


def format_graphql_enum_name(enum_name):
    """GraphQL has rules for Enum names, e.g., cannot begin
    with double underscores or contain special characters.
    """
    formatted_name = enum_name.replace(" ", "_").replace("-", "_")
    formatted_name = re.sub(r"^_+", "_", formatted_name)
    formatted_name = re.sub(r"\W+", "", formatted_name)
    return formatted_name



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from graphene import Field, InputObjectType
from djraphql.django_models import get_pk_fields_for_model
from graphql import Undefined


class PkInputType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)
        pk_fields = registry.get_pk_fields_for_model(model_class)

        if len(pk_fields) == 1:
            model_field = registry.get_model_fields(model_class)[pk_fields[0].name]
            return registry.get_graphene_type_or_default(model_field)

        return type(
            "{}PKInputType".format(model_class.__name__),
            (InputObjectType,),
            {
                field.name: Field(
                    registry.schema.type_map[field.get_internal_type()],
                    required=True,
                    default_value=Undefined,
                )
                for field in pk_fields
            },
        )



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from graphene import Field, ObjectType
from ...resolvers.field import build_field_resolver
from ...mappings import Types


specialized_field_types_by_aggregate_field_name = {
    "_count": "IntegerField",
    "_avg": "FloatField",
}


class AggregateBasicType(CacheableTypeBuilder, AbstractTypeBuilder):
    @classmethod
    def cache_key(cls, type_key, **kwargs):
        model_class = kwargs["model_class"]
        aggregate_field_name = kwargs["aggregate_field_name"]
        if (
            aggregate_field_name
            in specialized_field_types_by_aggregate_field_name.keys()
        ):
            return "{}/{}{}".format(
                model_class.__name__, type_key, aggregate_field_name
            )

        return "{}/{}".format(model_class.__name__, type_key)

    @staticmethod
    def make(registry, **kwargs):
        model_class = kwargs["model_class"]
        aggregate_field_name = kwargs["aggregate_field_name"]
        type_key = kwargs["type_key"]
        entity_class = registry.get_entity_class(model_class)
        aggregate_basic_type_attrs = {}

        # Edge-case for e.g. _count, where we need to use a different type
        # than the source column. So if we count(name), we want the result
        # to be an Int, not a String.
        registry_key = type_key
        if aggregate_field_name in specialized_field_types_by_aggregate_field_name:
            registry_key = "{}{}".format(registry_key, aggregate_field_name)

        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()

            field_name = field.name
            if field.concrete and not field.is_relation:
                type_name = (
                    specialized_field_types_by_aggregate_field_name.get(
                        aggregate_field_name
                    )
                    or field.get_internal_type()
                )
                graphene_type = registry.schema.FIELD_TYPE_MAP[type_name]
                arg = registry.get_or_create_type(
                    Types.WHERE_PREDICATE_INPUT_TYPE,
                    graphene_type=graphene_type,
                )
                aggregate_basic_type_attrs[field_name] = graphene_type(
                    args={"having": arg()}
                )
            elif field.concrete and (field.many_to_one or field.one_to_one):
                aggregate_basic_type_attrs[field_name] = Field(
                    registry.lambda_from_registry(field.related_model, registry_key),
                    required=not field.null,
                )
                aggregate_basic_type_attrs[
                    "resolve_{}".format(field_name)
                ] = build_field_resolver(field, registry)
                type_name = (
                    specialized_field_types_by_aggregate_field_name.get(
                        aggregate_field_name
                    )
                    or field.get_internal_type()
                )
                graphene_type = registry.schema.FIELD_TYPE_MAP[type_name]
                aggregate_basic_type_attrs["{}_id".format(field_name)] = graphene_type(
                    required=not field.null
                )
            elif field.many_to_one or field.one_to_one:
                aggregate_basic_type_attrs[field_name] = Field(
                    registry.lambda_from_registry(field.related_model, registry_key),
                    required=not field.field.null,
                )
                aggregate_basic_type_attrs[
                    "resolve_{}".format(field_name)
                ] = build_field_resolver(field, registry)
                type_name = field.field.get_internal_type()
                graphene_type = registry.schema.FIELD_TYPE_MAP[type_name]
                aggregate_basic_type_attrs["{}_id".format(field_name)] = graphene_type(
                    required=not field.field.null
                )
            elif field.one_to_many or field.many_to_many:
                aggregate_basic_type_attrs[field_name] = Field(
                    registry.lambda_from_registry(field.related_model, registry_key),
                )
                aggregate_basic_type_attrs[
                    "resolve_{}".format(field_name)
                ] = build_field_resolver(field, registry)

        result = type(
            "{}{}{}".format(
                model_class.__name__, "Aggregate", aggregate_field_name or ""
            ),
            (ObjectType,),
            aggregate_basic_type_attrs,
        )

        return result



import logging
import traceback
from graphene import (
    String,
    List,
    Boolean,
    Field,
    Argument,
    Mutation,
)
from graphene.types.generic import GenericScalar
from graphene.utils.str_converters import to_snake_case
from djraphql.access_permissions import UPDATE
from ...util import MutationAccumulator
from ...util.validation import FatalValidationError
from ....abstract_type_builder import MutationRootTypeBuilder, CacheableTypeBuilder
from ...mappings import Types


class UpdateMutation(CacheableTypeBuilder, MutationRootTypeBuilder):
    @classmethod
    def make(cls, registry, **kwargs):
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)
        update_input_type = registry.get_or_create_type(
            Types.UPDATE_INPUT_TYPE, model_class=model_class
        )
        pk_input_type = registry.get_or_create_type(
            Types.PK_INPUT_TYPE, model_class=model_class
        )
        model_name = model_class.__name__
        mutation_arguments_type = type(
            "{}UpdateArguments".format(model_name),
            (object,),
            {
                "pk": Argument(pk_input_type, required=True),
                "tag": Argument(String),
                "data": Argument(update_input_type, required=True),
            },
        )

        return type(
            "Update{}".format(model_name),
            (Mutation,),
            {
                "Arguments": mutation_arguments_type,
                "success": Boolean(),
                "pks_to_tags": List(registry.get_or_create_type(Types.TAG_TO_PK_TYPE)),
                "validation_errors": GenericScalar(),
                "result": Field(
                    registry.lambda_from_registry(model_class, Types.BASIC_TYPE)
                ),
                "mutate": cls._catch_all_exceptions(registry, model_class),
                "__doc__": cls.get_docs(model_class),
            },
        )

    @staticmethod
    def _catch_all_exceptions(registry, model_class):
        resolver = registry.get_or_create_type(
            Types.UPDATE_MUTATION_RESOLVER, model_class=model_class
        )

        def _catch_all(*args, **kwargs):
            accumulator = MutationAccumulator()
            try:
                return resolver(*args, accumulator=accumulator, **kwargs)
            except FatalValidationError as ve:
                mutation_type = registry.get_or_create_type(
                    Types.UPDATE_MUTATION_TYPE, model_class=model_class
                )
                return mutation_type(
                    success=False,
                    pks_to_tags=accumulator.pks_to_tags,
                    validation_errors=accumulator.validation_errors,
                    result=accumulator.root_result,
                )
            except:
                logging.error(
                    "Error resolving object at path {}".format(accumulator.current_path)
                )
                traceback.print_exc()
                raise

        return _catch_all

    @staticmethod
    def get_field_name(model_class):
        return "update_{}".format(to_snake_case(model_class.__name__))

    @staticmethod
    def get_docs(model_class):
        return """Given a {} identifier (pk) and an object representing the
            changed values (data), updates an existing object in the database.
        """.format(
            model_class.__name__
        )

    @staticmethod
    def get_required_access_permissions():
        return UPDATE



from ..util.visitors import visit_query


def build_by_pk_resolver(model_class, registry):
    entity_class = registry.get_entity_class(model_class)

    def by_pk_resolver(parent, info, **kwargs):
        queryset = entity_class.get_queryset(info.context).filter(**kwargs)
        artifacts = visit_query(registry, info)
        optimizer = artifacts.get("queryset_optimizer")
        if not optimizer:
            return queryset
        return optimizer.optimize(queryset).get()

    return by_pk_resolver



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from ...mappings import Types
from graphene import Field, List, ObjectType
from ...resolvers.field import build_field_resolver
from ...resolvers.list import build_list_resolver
from ...util.arguments import create_query_arguments
from .....django_models import get_field_by_name, get_target_field


class BasicType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        basic_type_attrs = {}
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)

        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()

            if (
                field.is_relation
                and not registry.get_entity_class(field.related_model).is_readable()
            ):
                continue

            # Handle concrete, non-relation fields
            if not field.is_relation:
                graphene_type = registry.get_graphene_type_or_default(model_field)

                # Enum/choice fields are handled as special cases
                if field.choices:
                    graphene_type = registry.get_or_create_type(
                        Types.ENUM_TYPE,
                        model_class=model_class,
                        field=field,
                        choices=field.choices,
                    )

                basic_type_attrs[field.name] = Field(
                    graphene_type, required=not field.null
                )
                continue

            # Handle *-to-one relational fields (which may be concrete)
            if field.many_to_one or field.one_to_one:
                basic_type_attrs[field.name] = Field(
                    registry.lambda_from_registry(
                        field.related_model, Types.BASIC_TYPE
                    ),
                    required=not field.null if field.concrete else False,
                )
                basic_type_attrs[
                    "resolve_{}".format(field.name)
                ] = build_field_resolver(field, registry)

                # If this relational field is concrete, also add a *_id property
                # to the GraphQL type. I.e., a foo that has a bar will have both
                # barId and bar { id } available on the type.
                if field.concrete:
                    reverse_model_field = registry.get_model_fields(
                        field.related_model
                    )[get_target_field(field).name]

                    graphene_type = registry.get_graphene_type_or_default(
                        reverse_model_field
                    )
                    basic_type_attrs["{}_id".format(field.name)] = graphene_type(
                        required=not field.null
                    )
                continue

            # Handle *-to-many relational fields (which are never concrete)
            if field.one_to_many or field.many_to_many:
                where_type = registry.get_or_create_type(
                    Types.WHERE_CLAUSE_TYPE, model_class=field.related_model
                )
                order_by_type = registry.get_or_create_type(
                    Types.ORDER_CLAUSE_TYPE, model_class=field.related_model
                )
                basic_type_attrs[field.name] = List(
                    registry.lambda_from_registry(
                        field.related_model, Types.BASIC_TYPE
                    ),
                    args=create_query_arguments(where_type, order_by_type),
                )

                basic_type_attrs["resolve_{}".format(field.name)] = build_list_resolver(
                    field.related_model,
                    registry,
                    related_name=(
                        field.name if field.many_to_many else field.related_name
                    ),
                )
                continue

        # Add computed (property) fields defined on the Entity class
        for computed_field in entity_class._get_computed_fields():
            basic_type_attrs[computed_field.name] = computed_field.graphene_type()
            basic_type_attrs[
                f"resolve_{computed_field.name}"
            ] = build_computed_field_resolver(entity_class, computed_field)

        return type(
            model_class.__name__,
            (ObjectType,),
            basic_type_attrs,
        )


def build_computed_field_resolver(entity_class, computed_field):
    resolver = computed_field.get_resolver(entity_class)

    def resolve(parent, info, **kwargs):
        return resolver(info.context, parent, **kwargs)

    return resolve



from six import iteritems
from graphql.language.visitor import Visitor
from graphene.utils.str_converters import to_snake_case
from djraphql.django_models import get_field_by_name
from graphql.execution.values import get_argument_values

IGNORED_FIELDS = set(["__typename"])


class QueryFieldsVisitor(Visitor):
    def __init__(self, registry, info):
        super().__init__()
        self.registry = registry
        self.type_lookup = registry._node_graphql_type_name_to_model_class
        self.info = info
        self.graphql_type_stack = [info.parent_type]
        self.model_class_stack = []

    def _get_terminal_gql_type(self, gql_type):
        while hasattr(gql_type, "of_type"):
            gql_type = gql_type.of_type
        return gql_type

    def enter_field(
        self,
        node,
        key,
        parent,
        path,
        ancestors,
    ):
        # Get the parent type of this field
        parent_gql_type = self.graphql_type_stack[-1]

        # Get a dictionary of the arguments associated with this field
        field_def = parent_gql_type.fields.get(node.name.value)

        # If the query contains GraphQL-spec-level fields like __typename,
        # field_def will be None here. We can ignore when this happens.
        if field_def is None and node.name.value in IGNORED_FIELDS:
            return

        arguments = get_argument_values(
            field_def,
            node,
            self.info.variable_values,
        )

        # Do we call hello_field?
        if parent_gql_type.name in self.type_lookup:
            model_class = self.type_lookup[parent_gql_type.name]["model_class"]
            field = try_get_related_field(model_class, node.name.value)
            if field:
                self.hello_field(field, arguments, graphql_node=node)
            else:
                # The associated model doesn't have a field with the same name.
                # The only possibility is that this field corresponds to a ComputedField.
                # If this ComputedField defined a depends_on list, we process it here.
                entity_class = self.type_lookup[parent_gql_type.name]["entity_class"]
                for computed_field in entity_class._get_computed_fields():
                    for dependency in computed_field.depends_on:
                        field = try_get_related_field(model_class, dependency)
                        if field:
                            self.hello_field(field, arguments, graphql_node=node)
                            self.goodbye_field(field)

        # If we're traversing a relationship (e.g., we're entering the
        # "artists" node within the "LabelsMany" query), we need to push the new
        # GraphQLTypeObject class and model class onto their respective stacks.
        if node.selection_set:
            maybe_nonterminal_gql_type = parent_gql_type.fields.get(
                node.name.value
            ).type
            next_gql_type = self._get_terminal_gql_type(maybe_nonterminal_gql_type)
            self.graphql_type_stack.append(next_gql_type)

            if next_gql_type.name in self.type_lookup:
                model_class = self.type_lookup[next_gql_type.name]["model_class"]
                if not self.model_class_stack:
                    self.hello_optimizable_tree(model_class, arguments)
                self.model_class_stack.append(model_class)

    def hello_optimizable_tree(self, model_class, arguments):
        pass

    def goodbye_optimizable_tree(self, model_class):
        pass

    def hello_field(self, field, arguments, graphql_node=None):
        pass

    def goodbye_field(self, field):
        pass

    def leave_field(
        self,
        node,
        key,
        parent,
        path,
        ancestors,
    ):
        # If we're traversing a relationship, e.g. we're leaving the "artists" node
        # within the "LabelsMany" query, we need to pop our GraphQLType stack
        # and _possibly_ our model stack.
        if node.selection_set:
            self.graphql_type_stack.pop()
            if len(self.model_class_stack) == 1:
                # We're about to pop the last item from our model stack,
                # that means we're exiting the optimizable tree, so call that handler.
                self.goodbye_optimizable_tree(self.model_class_stack.pop())
            elif self.model_class_stack:
                # There will still be items left in our model stack, we're still
                # within the optimizable tree. In this case, call goodbye_field.
                self.model_class_stack.pop()
                field = try_get_related_field(
                    self.model_class_stack[-1], node.name.value
                )
                if field:
                    self.goodbye_field(field)
            return

        # We are not traversing a relationship. If we have a non-empty
        # model stack, we need to call the goodbye_field handler.
        if self.model_class_stack:
            field = try_get_related_field(self.model_class_stack[-1], node.name.value)
            if field:
                self.goodbye_field(field)


def try_get_related_field(model_class, selection_name):
    names_to_try = [
        # First try just getting the field from the model via the selection's name
        # in the GraphQL query. This is usually a camelCased name. We do this in
        # case the user has defined their model's related_name(s)
        # in camelCase instead of snake_case
        selection_name,
        # Next, attempt to get the field by snake_casing the name of the selection.
        to_snake_case(selection_name),
    ]

    for name in names_to_try:
        try:
            return get_field_by_name(model_class, name)
        except:  # FieldDoesNotExist
            pass



def build_field_resolver(field, registry):
    entity_class = registry.get_entity_class(field.related_model)

    def resolve_field(parent, info, **kwargs):
        return getattr(parent, field.name, None)

    return resolve_field



from .query_fields_visitor import QueryFieldsVisitor
from ..queryset_optimizer import QuerysetOptimizer


class NodeInfo(object):
    def __init__(self, key, model_class, field_names, arguments=None, alias=None):
        self.key = key
        self.model_class = model_class
        self.field_names = field_names
        self.select_related_keys = []
        self.arguments = arguments
        self.alias = alias
        self.alias_path = None


class QuerysetOptimizationVisitor(QueryFieldsVisitor):
    """Problem: We allow a caller to specify e.g. a 'where' filter in a
    GraphQL query at any level of the query: root, child, grandchild, ...
    When there is a 'where' at a non-root level, we will fall victim
    to the N+1 problem and issue as many queries as there are children.
    Solution: To avoid N+1 queries, we traverse the GraphQL query's AST
    to find opportunities to use Django's Prefetch objects.
    Example: consider the following query
    query {
      albums(where: {author: _eq: "bob"}}) {
        name
        releaseDate
        songs(where: {starCount: {_eq: 5}}) {
          id
          name
          starredBy {
            name
          }
        }
      }
    }
    If bob has released 5 albums, and on each of those albums, 3 songs
    have 5 stars, then:
    1x query for albums where bob is the author... returns 5 albums
    5x (5 albums) query for 5-star songs on each album... each query returns 3 songs
    15x (3 songs x 5 albums) query for the names of the people that
    starred the song.
    For a grand total of 75 queries to the database in a single request!
    Using Prefetch can bring that down the number of level deep
    the query goes, which in this case is 3.

    This class does two things to optimize the SQL executed for this query;
    (a) Update the queryset to only fetch the columns necessary to fulfill the request
    (b) Update the queryset to prefetch/select relate fields to avoid N+1
    """

    def __init__(self, registry, info):
        self.prefetch_stack = []
        self.prefetch_info_by_key = {}
        super(QuerysetOptimizationVisitor, self).__init__(registry, info)

    def hello_field(self, field, arguments, graphql_node=None):

        if field.is_relation:
            # Subtle thing here to account for how ORM optimizes query prefetching.
            # If the field is concrete, we need to add the field's name to the
            # field names for the node we're leaving.
            # If not concrete we need to add the reverse related name to the
            # field names for the node we're entering.
            field_names = []
            if field.concrete:
                self.prefetch_stack[-1].field_names.append(field.name)
            else:
                field_names = [field.field.name]

            # Push a NodeInfo onto the stack.
            alias = graphql_node.alias.value if graphql_node.alias else None
            self.prefetch_stack.append(
                NodeInfo(field.name, field.related_model, field_names, arguments, alias)
            )

        else:
            # This is just a scalar, which we add to the top of the stack's
            # list of field names to be passed to .only() later.
            self.prefetch_stack[-1].field_names.append(field.name)

    def goodbye_field(self, field):

        # If we aren't leaving a relational field, there's nothing to do
        if not field.is_relation:
            return

        if field.one_to_many or field.many_to_many:
            # *-to-many relations are handled via a later prefetch_related,
            # so we need to perform some bookkeeping by updating prefetch_info_by_key.

            # Fun edge case for handling aliased queries on the same field, e.g.
            # LabelByPk(pk: $labelId) {
            #   artists(where: {id: {_eq: $artistId}}) {
            #     name
            #     albums(where: {id: {_eq: $albumId}}) { title }
            #     allAlbums: albums(orderBy: {title: desc}) { title }
            #   }
            #   allArtists: artists(orderBy: {name: desc}) {
            #     name
            #     albums(where: {id: {_eq: 2}}) { title }
            #     allAlbums: albums(orderBy: {title: desc}) { title }
            #   }
            # }
            # Django handles this use case via the "to_attr" kwarg to Prefetch.
            # The following code constructs the path that works w/ aliases / to_attr.

            # We construct a path where aliases are _not_ used, because it allows
            # us to later determine if we _really_ need to use to_attr, which loads
            # its list into memory. We want to avoid doing this if e.g. a single
            # query happened to use an alias, and only use to_attr when two or more
            # queries of the same field exist within the same node.
            key_path = "__".join([pi.key for pi in self.prefetch_stack if pi.key])
            if key_path not in self.prefetch_info_by_key:
                self.prefetch_info_by_key[key_path] = []

            # Add to the list of prefetches on this key path. If the list on
            # this key path has more than 1 item, then we know we must use to_attr.
            prefetch_info = self.prefetch_stack.pop()
            self.prefetch_info_by_key[key_path].append(prefetch_info)

            # Set the correct path to be passed to Prefetch: all nonterminal
            # parts of the path use its alias if one exists, else its normal key.
            # The terminal/last part always uses its normal key.
            path_parts = [pi.alias or pi.key for pi in self.prefetch_stack if pi.key]
            path_parts.append(prefetch_info.key)
            prefetch_info.alias_path = "__".join(path_parts)
        elif field.one_to_one or field.many_to_one:
            # *-to-one relations are handled via a later select_related,
            # we've got bookkeeping here as well, but it's a bit more complex.
            # We must collapse/coalesce the top 2 items in the prefetch_stack.
            to_one = self.prefetch_stack.pop()

            # Add the popped item's key to the new top item's select_related_keys list.
            self.prefetch_stack[-1].select_related_keys.append(to_one.key)

            # Add the popped item's select_related_keys to the new top item's
            # select_related_keys, each prefixed with the popped item's key.
            for select_related_key in to_one.select_related_keys:
                self.prefetch_stack[-1].select_related_keys.append(
                    "{}__{}".format(to_one.key, select_related_key)
                )

            # Add the popped item's field_names to the new top item's field_names,
            # but prefixed with the popped item's key.
            for field_name in to_one.field_names:
                self.prefetch_stack[-1].field_names.append(
                    "{}__{}".format(to_one.key, field_name)
                )

    def hello_optimizable_tree(self, model_class, arguments):
        self.prefetch_stack.append(NodeInfo(None, model_class, [], arguments))

    def goodbye_optimizable_tree(self, model_class):
        self.artifacts = {
            "queryset_optimizer": QuerysetOptimizer(
                self.prefetch_stack.pop(), self.prefetch_info_by_key
            )
        }



from djraphql.schema_builder import SchemaBuilder
from djraphql.entities import Entity
from djraphql.fields import AllModelFields, ModelField, ComputedField
from djraphql.filters import FilterBackend
from djraphql.access_permissions import (
    C,
    R,
    U,
    D,
    Create,
    Read,
    Update,
    Delete,
    RequestAuthorizationCheck,
)



from functools import reduce
from graphene.utils.str_converters import to_camel_case


class FatalValidationError(Exception):
    pass


class ValidatableInput:
    def __init__(self, raw_input, root_model_class, is_insert=True):
        self._raw_input = raw_input
        self._root_model_class = root_model_class
        self._is_insert = is_insert

    def to_value(self, camel_case_keys=False, lift_lists=True):

        value = self._raw_input
        if lift_lists:
            # The input format for inserts and updates is slightly different:
            # Update input objects go one level deeper -- via an "items" key -- to get
            # the actual item being updated. So to lift the *-to-many fields up,
            # we need to account for that. This extract_list helper performs the
            # correct traversal based on the path variable.
            path = [] if self._is_insert else ["items"]
            extract_list = lambda items: reduce(lambda a, b: a[b], path, items)
            value = self._lift_lists_to_related_key(
                value, self._root_model_class, extract_list
            )

        if camel_case_keys:
            value = self._camel_case_keys(value)

        return value

    @classmethod
    def _camel_case_keys(cls, value):
        if not value or not isinstance(value, dict):
            return value

        result = {}
        for k in value:
            result[to_camel_case(k)] = cls._camel_case_keys(value[k])

        return result

    @classmethod
    def _lift_lists_to_related_key(cls, value, model_class, extract_list):
        """Dealing with the raw input object is cumbersome due to the intermediate
        layers of the object tree that contain things like updateStrategy, etc.

        This method provides a way for the consumer of the library to instead receive
        a structure that mirrors the model structure, which is more convenient for
        validation.

        Args:
            value: object. e.g. for an updateLabel, it may look like:
                {
                    "artists": {
                        "updateStrategy": "replace",
                        "items: [
                            {"pk": 1, "data": {"name": "Jim"}},
                            {"data": {"name": "Bob"}}
                        ]
                    }
                }
            model_class: the model_class being inserted/updated
            extract_list: helper function to get to the list of items
        Returns:
            object: example based on above value argument,
                {
                    "artists": [
                        {"id": 1, "name": "Jim"},
                        {"name": "Bob"}
                    ]
                }
        """

        # Iterate through the model's to-many fields
        for f in model_class._meta.get_fields():
            if (
                not f.name in value
                or not f.is_relation
                or f.one_to_one
                or f.many_to_one
            ):
                continue

            results = []
            for item in extract_list(value[f.name]):
                new_item = cls._lift_lists_to_related_key(
                    item["data"].copy(), f.related_model, extract_list
                )
                if "pk" in item:
                    new_item["pk"] = item["pk"]
                results.append(new_item)

            value[f.name] = results

        return value


def perform_validation(
    entity_class, raw_input_value, model_class, parent, info, is_insert=False
):
    if not entity_class.validator:
        return (True, None)

    validatable_input = ValidatableInput(
        raw_input_value, model_class, is_insert=is_insert
    )

    return entity_class.validator.validate(validatable_input, parent=parent, info=info)



import copy


class MutationAccumulator:
    def __init__(self):
        self.current_path = []
        self.pks_to_tags = []
        self.validation_errors = []  # e.g. [{"path": [], "errors": {"foo": "bar"}}]
        self.fatal_error = None
        self.root_result = None

    def push_path_part(self, path_part):
        self.current_path.append(path_part)

    def pop_path_part(self):
        return self.current_path.pop()

    def add_validation_error(self, error):
        self.validation_errors.append(
            {
                "path": copy.copy(self.current_path),
                "errors": error,
            }
        )

    def add_pk_to_tag(self, pk, tag, _type):
        self.pks_to_tags.append({"pk": pk, "tag": tag, "type": _type})

    def set_fatal_error(self, error):
        self.fatal_error = error

    def should_halt_execution(self):
        return bool(self.fatal_error)



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from graphene import InputObjectType, Enum
from graphql import Undefined


class SortDirection(Enum):
    asc = "ascending"
    desc = "descending"


class OrderClause(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        # Build where predicate InputObjects to allow filtering in queries.
        order_by_type_attrs = {}
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)

        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()
            if not field.concrete:
                continue

            # Convert names of ForeignKey/OneToOneFields from 'foo' to 'foo_id'
            field_name = field.name
            if field.many_to_one or field.one_to_one and field_name != "id":
                field_name = "{}_id".format(field_name)
            order_by_type_attrs[field_name] = SortDirection(default_value=Undefined)

        return type(
            "OrderBy{}".format(model_class.__name__),
            (InputObjectType,),
            order_by_type_attrs,
        )



from ....abstract_type_builder import AbstractTypeBuilder
from graphene import String, ObjectType


class TagToPkType(AbstractTypeBuilder):
    @classmethod
    def cache_key(cls, type_key, **kwargs):
        return type_key

    @staticmethod
    def make(registry, **kwargs):
        return type(
            "TagPrimaryKeyMutationMap",
            (ObjectType,),
            {
                "tag": String(),
                "pk": String(),
                "type": String(),
            },
        )



from graphene import Boolean, Argument, Mutation
from graphene.utils.str_converters import to_snake_case
from djraphql.access_permissions import DELETE
from ....abstract_type_builder import MutationRootTypeBuilder, CacheableTypeBuilder
from ...mappings import Types


class DeleteMutation(CacheableTypeBuilder, MutationRootTypeBuilder):
    @classmethod
    def make(cls, registry, **kwargs):
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)
        model_name = model_class.__name__
        pk_input_type = registry.get_or_create_type(
            Types.PK_INPUT_TYPE, model_class=model_class
        )
        mutation_arguments_type = type(
            "{}DeleteArguments".format(model_name),
            (object,),
            {"pk": Argument(pk_input_type, required=True)},
        )

        return type(
            "Delete{}".format(model_name),
            (Mutation,),
            {
                "Arguments": mutation_arguments_type,
                "success": Boolean(),
                "mutate": registry.get_or_create_type(
                    Types.DELETE_MUTATION_RESOLVER, model_class=model_class
                ),
                "__doc__": cls.get_docs(model_class),
            },
        )

    @staticmethod
    def get_field_name(model_class):
        model_name = model_class.__name__
        return "delete_{}".format(to_snake_case(model_name))

    @classmethod
    def get_docs(cls, model_class):
        model_name = model_class.__name__
        return """Given a {} identifier (pk), dangerously deletes the
            associated object from the database.
        """.format(
            model_name
        )

    @staticmethod
    def get_required_access_permissions():
        return DELETE



from six import iteritems
from functools import reduce
from graphene.utils.str_converters import to_snake_case
from ..util.arguments import update_queryset_based_on_arguments

IGNORED_FIELDS = set(["__typename"])


def build_aggregate_resolver(
    model_class, registry, aggregate_field_names_to_django_functions, related_name=None
):
    entity_class = registry.get_entity_class(model_class)

    def aggregate_resolver(parent, info, **kwargs):

        # Leaving this here as the QueryFieldsVisitor class doesn't yet
        # handle keywords like _sum, _max, _avg, etc. and thus the mechanism
        # that blocks unauthorized access *may* not work, so purely as a
        # fail-safe, this will stay for now. The aggregate querying
        # functionality needs a total rewrite/rethink anyhow.
        if not entity_class.is_readable_by_context(info.context):
            raise Exception("Invalid permissions.")

        # If related_name and parent are not None, then we are resolving a
        # subselect, which has already been prefetched via the code below.
        if related_name and parent is not None:
            return getattr(parent, related_name).all()

        # Call any filter_backends that were provided to us.
        queryset = entity_class.get_queryset(info.context)

        # For aggregate queries, all non-aggregate fields are included in the
        # resulting SQL query's group-by clause.
        grouped_fields, aggregated_fields_by_fn = _parse_aggregate_query(
            model_class,
            info,
            aggregate_field_names_to_django_functions.keys(),
        )

        # Filter the root queryset based on the arguments in the query
        queryset = update_queryset_based_on_arguments(
            queryset, kwargs, take_slice=False
        )

        # If there are no group-by fields, we simply need to call .aggregate()
        # with the right column and aggregation-function.
        # This does issue a DB query for every aggregation :(
        if not grouped_fields:
            results = []
            for aggregate_field_type, aggregated_fields in iteritems(
                aggregated_fields_by_fn
            ):
                fn = aggregate_field_names_to_django_functions[aggregate_field_type]
                for aggregated_field in aggregated_fields:
                    aggregate_result = queryset.aggregate(fn(aggregated_field))
                    value = aggregate_result[
                        "{}_{}".format(aggregated_field, aggregate_field_type)
                    ]
                    results.append({aggregate_field_type: {aggregated_field: value}})

            return results

        # We do have a group-by clause, so we have to do a bit more work...
        # Set the values to be in the GROUP BY clause
        queryset = queryset.values(*grouped_fields)

        # Use the mapping to make the annotations on the queryset
        labeled_aggregated_fields_by_aggregate_type = {}
        for aggregate_field_type, aggregated_fields in iteritems(
            aggregated_fields_by_fn
        ):
            labeled_aggregated_fields_by_aggregate_type[aggregate_field_type] = {}

            # fn here will be something like django.models.Sum
            fn = aggregate_field_names_to_django_functions[aggregate_field_type]

            # Group the aggregated fields by their aggregation type
            annotations = {
                "{}__{}".format(aggregate_field_type, aggregated_field): fn(
                    aggregated_field
                )
                for aggregated_field in aggregated_fields
            }

            # Annotate the queryset with each aggregation
            queryset = queryset.annotate(**annotations)

            # Optimization: repurpose our annotations dictionary to aid in
            # our transforming the resultset from the DB into the object shape
            # expected by the GraphQL resolve machinery by creating a list of
            # our nested references and reversing it, e.g.
            # artists__albums__star_count -> ['star_count', 'albums', 'artists']
            # Which is then used in the "reduce" call below to build a result
            # dictionary from the bottom-up.
            for aggregated_field_label, _ in iteritems(annotations):
                labels = aggregated_field_label.split("__")
                labels.reverse()
                labeled_aggregated_fields_by_aggregate_type[
                    aggregate_field_type
                ].update({aggregated_field_label: labels})

        # Do the same optimization we do for grouped fields that we did for
        # aggregated fields (see above).
        grouped_fields_by_resultset_field = {}
        for grouped_field in grouped_fields:
            split_fields = grouped_field.split("__")
            split_fields.reverse()
            grouped_fields_by_resultset_field[grouped_field] = split_fields

        # We must map the SQL result set to the shape of the GraphQL request
        results = []
        for row in queryset:
            # Grouped subqueries
            for resultset_name, split_fields in iteritems(
                grouped_fields_by_resultset_field
            ):
                group_nested = reduce(
                    lambda a, b: {b: a}, [row[resultset_name]] + split_fields
                )
                row.update(group_nested)

            # Aggregate subqueries
            for aggregate_field_name, labeled_aggregated_fields in iteritems(
                labeled_aggregated_fields_by_aggregate_type
            ):
                row[aggregate_field_name] = {}
                for label, split_and_reversed_label in iteritems(
                    labeled_aggregated_fields
                ):
                    nested = reduce(
                        lambda a, b: {b: a}, [row[label]] + split_and_reversed_label
                    )
                    row[aggregate_field_name].update(nested[aggregate_field_name])
            results.append(row)
        return results

    return aggregate_resolver


# TODO: This function should use a visitor, like how query-optimization works.
def _parse_aggregate_query(model_class, info, aggregate_properties):
    def _inner_parse_aggregate_query_aggregate_selection(
        aggregate_fn, selections, parent_key=None
    ):
        result = []
        for selection in selections:
            if selection.name.value in IGNORED_FIELDS:
                continue

            if selection.selection_set and selection.selection_set.selections:
                result += _inner_parse_aggregate_query_aggregate_selection(
                    aggregate_fn,
                    selection.selection_set.selections,
                    parent_key=(
                        "{}__{}".format(parent_key, to_snake_case(selection.name.value))
                        if parent_key
                        else to_snake_case(selection.name.value)
                    ),
                )
            elif parent_key:
                result.append(
                    "{}__{}".format(parent_key, to_snake_case(selection.name.value))
                )
            else:
                result.append(to_snake_case(selection.name.value))

        return result

    def _inner_parse_aggregate_query(
        model_class, parent_type, selections, current_key=None
    ):
        grouped_fields = []
        aggregated_fields_by_fn = {}
        for selection in selections:
            if selection.name.value in IGNORED_FIELDS:
                continue

            if selection.name.value in aggregate_properties:
                if selection.name.value not in aggregated_fields_by_fn:
                    aggregated_fields_by_fn[selection.name.value] = []

                aggregated = _inner_parse_aggregate_query_aggregate_selection(
                    selection.name.value, selection.selection_set.selections
                )
                aggregated_fields_by_fn[selection.name.value] += aggregated
            elif selection.selection_set and selection.selection_set.selections:
                grouped, _ = _inner_parse_aggregate_query(
                    model_class,
                    parent_type,
                    selection.selection_set.selections,
                    current_key=selection.name.value,
                )
                grouped_fields.append(
                    to_snake_case(selection.name.value) + "__" + "__".join(grouped)
                )
            else:
                grouped_fields.append(to_snake_case(selection.name.value))

        return grouped_fields, aggregated_fields_by_fn

    parent_type = info.parent_type.fields.get(info.field_name).type
    selections = info.field_nodes[0].selection_set.selections
    return _inner_parse_aggregate_query(
        model_class,
        getattr(parent_type, "of_type", parent_type),
        selections,
    )



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from ...mappings import Types
from graphene import Field, ObjectType
from ...resolvers.group_by import build_group_by_resolver


class AggregateGroupByType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        attrs = {}
        model_class = kwargs["model_class"]
        entity_class = registry.get_entity_class(model_class)

        for model_field in registry.get_model_fields(model_class).values():
            field = model_field.get_django_model_field()

            if field.concrete and not field.is_relation:
                attrs[field.name] = registry.get_graphene_type_or_default(model_field)()
            elif field.concrete and field.is_relation:
                # Example: model_class is Album, and field is Album.artist
                attrs[field.name] = Field(
                    registry.lambda_from_registry(
                        field.related_model, Types.AGGREGATE_GROUP_BY_TYPE
                    )
                )
                attrs["resolve_{}".format(field.name)] = build_group_by_resolver(field)
            elif field.one_to_many:
                attrs[field.name] = Field(
                    registry.lambda_from_registry(
                        field.related_model, Types.AGGREGATE_GROUP_BY_TYPE
                    )
                )
                attrs["resolve_{}".format(field.name)] = build_group_by_resolver(field)

        return type(
            "{}GroupBy".format(model_class.__name__),
            (ObjectType,),
            attrs,
        )



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from ...mappings import Types
from graphene import Field, InputObjectType, String
from graphql import Undefined


class NestedItemInsertInputType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        input_class_attrs = {}
        model_class = kwargs["model_class"]

        return type(
            "NestedItemInsert{}".format(model_class.__name__),
            (InputObjectType,),
            {
                "tag": Field(String, default_value=Undefined),
                "data": Field(
                    registry.lambda_from_registry(model_class, Types.INSERT_INPUT_TYPE),
                    required=True,
                    default_value=Undefined,
                ),
            },
        )



from .query_fields_visitor import QueryFieldsVisitor


class ReadPermissionsCheckVisitor(QueryFieldsVisitor):
    def hello_field(self, field, arguments, graphql_node=None):
        # If this is the "fooId" field (vs. the "foo" field), then
        # it's perfectly fine to read, because it's a scalar field
        # on a type that's already passed permissions-check successfully.
        # Thus, only check permissions if the GraphQL node has subselections.
        if graphql_node.selection_set:
            self._check_permissions(field.related_model)

    def hello_optimizable_tree(self, model_class, arguments):
        # Check permissions for reading the _root-level_ object.
        self._check_permissions(model_class)

    def _check_permissions(self, model_class):
        entity_class = self.registry.get_entity_class(model_class)
        if not entity_class.is_readable_by_context(self.info.context):
            raise Exception("Invalid permissions.")



def get_pk_fields_for_model(model_class, available_model_classes):
    """Get all fields comprising a model class's primary key.
    """
    return tuple(
        [
            f
            for f in model_class._meta.get_fields()
            if getattr(f, "primary_key", False)
            and (not f.is_relation or f.related_model in available_model_classes)
        ]
    )


def get_field_by_name(model_class, field_name):
    """Given a Django model class and a field name, return the field.
    """
    if hasattr(model_class._meta, "get_field"):
        return model_class._meta.get_field(field_name)

    raise Exception("Cannot fetch {} field {}.".format(model_class, field_name))


def get_target_field(relational_field):
    """Given a relational Field, e.g. <ManyToOneRel: sample_music_app.artist>,
    gets the corresponding field on the referenced model.
    """

    # Django 2+
    if hasattr(relational_field, "target_field"):
        return relational_field.target_field

    # Django 1.8+
    if hasattr(relational_field, "related_field"):
        return relational_field.related_field

    raise Exception("Cannot get target field on field {}.".format(relational_field))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = "0.2.8"

setuptools.setup(
    name="djraphql",
    version=VERSION,
    author="Joel Gardner",
    author_email="joel@simondata.com",
    description="DjraphQL builds a flexible & performant GraphQL schema by examining your Django models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Radico/djraphql",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "django",
        "graphene",
        "six",
    ],
    pypi={"name": "djraphql", "version": VERSION},
)



from ....abstract_type_builder import AbstractTypeBuilder, CacheableTypeBuilder
from ...mappings import Types
from graphene import Enum, InputObjectType, List


class UpdateStrategy(Enum):
    merge = "merge"
    replace = "replace"


class ListUpdateInputType(CacheableTypeBuilder, AbstractTypeBuilder):
    @staticmethod
    def make(registry, **kwargs):
        """Generates a Graphene InputObjectType for the associated model.
        https://docs.graphene-python.org/en/latest/types/mutations/#inputfields-and-inputobjecttypes
        """
        input_class_attrs = {}
        model_class = kwargs["model_class"]
        nested_item_update_input_type = registry.get_or_create_type(
            Types.NESTED_ITEM_UPDATE_TYPE, model_class=model_class
        )
        return type(
            "{}ListUpdateInput".format(model_class.__name__),
            (InputObjectType,),
            {
                "update_strategy": UpdateStrategy(required=True, default_value="merge"),
                "items": List(nested_item_update_input_type, required=True),
            },
        )



from ....abstract_type_builder import AbstractTypeBuilder
from graphene import Boolean, InputObjectType, List
from graphql import Undefined


class WherePredicate(AbstractTypeBuilder):
    @classmethod
    def cache_key(cls, type_key, **kwargs):
        graphene_type = kwargs["graphene_type"]
        return graphene_type.__name__

    @staticmethod
    def make(registry, **kwargs):
        graphene_type = kwargs["graphene_type"]
        return type(
            "WherePredicate{}".format(graphene_type.__name__),
            (InputObjectType,),
            {
                "_eq": graphene_type(name="_eq", default_value=Undefined),
                "_neq": graphene_type(name="_neq", default_value=Undefined),
                "_gt": graphene_type(name="_gt", default_value=Undefined),
                "_gte": graphene_type(name="_gte", default_value=Undefined),
                "_lt": graphene_type(name="_lt", default_value=Undefined),
                "_lte": graphene_type(name="_lte", default_value=Undefined),
                "_in": List(graphene_type, name="_in", default_value=Undefined),
                "_nin": List(graphene_type, name="_nin", default_value=Undefined),
                "_is_null": Boolean(name="_is_null", default_value=Undefined),
                "_like": graphene_type(name="_like", default_value=Undefined),
                "_ilike": graphene_type(name="_ilike", default_value=Undefined),
            },
        )



from djraphql.django_models import get_pk_fields_for_model


class TypeRegistry:
    """The TypeRegistry is the control-center of DjraphQL. It takes a schema object
    which is created by the library user and exposes methods that are called during
    GraphQL-schema build-time and query resolution time.

    It takes a list of Entity classes (via register_entity_classes) and makes them
    accessible to the classes that build the Graphene types that eventually become
    GraphQL classes.

    Each builder type (e.g. basic.py) exposes a make method that takes a TypeRegistry
    instance as its first argument. Through this instance, type builders reference any
    related types associated with the model being processed.
    """

    _ENTITY_CLASS_KEY = "__entity_class__"
    _MODEL_FIELDS_KEY = "__model_fields_by_name__"

    def __init__(self, schema):
        """
        Args:
            schema: Instance of a class that implements
                    djraphql.schemas.abstract_type_builder.AbstractSchema.
        """
        self.schema = schema
        self._node_graphql_type_name_to_model_class = {}
        self._cache = {
            "models": {},
            "types": {},
        }

    def get_available_model_classes(self):
        """A convenience method that returns a list of all Model classes that
        should be exposed in the GraphQL API. Which is all the models that have
        an associated Entity class that was passed to register_entity_classes.
        """
        if not hasattr(self, "_available_model_classes"):
            self._get_available_model_classes = self._model_cache.keys()

        return self._get_available_model_classes

    @property
    def _model_cache(self):
        """A dictionary that stores any models whose Entity class we're aware of."""
        return self._cache["models"]

    @property
    def _type_cache(self):
        """A dictionary that stores any Graphene types we've created."""
        return self._cache["types"]

    def get_entity_class(self, model_class):
        return self._model_cache.get(model_class, {}).get(self._ENTITY_CLASS_KEY)

    def lambda_from_registry(self, model_class, key):
        """A heavily used method that returns a lambda that returns a Graphene type.
        This is necessary because circular relationships exist. If we're building a
        schema that contains a Foo model that is one-to-many with a Bar model, we'll
        start creating the Foo, notice that it has a field "bars" and then attempt
        to create the Bar type. But then we'll run into the Bar.foo field, and attempt
        to create the Foo type, creating an infinite loop.

        Graphene solves this by allowing the "type" argument on the Field be a lambda
        that returns a type, which allows us to avoid the infinite loop by providing
        a lambda (that returns the type) instead of the type itself.

        Args:
            model_class: class of the Django modele for the Graphene type we want
            key: string from the djraphql.schemas.default.mappings.Types class.
        Returns:
            A lambda that returns the Graphene type corresponding to the inputs.
        """

        def _get_type_from_registry():
            cache_key = "{}/{}".format(model_class.__name__, key)
            if cache_key not in self._type_cache:
                raise RuntimeError(
                    "{} does not exist in DjraphQL type cache.".format(cache_key)
                )

            return self._type_cache[cache_key]

        return _get_type_from_registry

    def register_entity_classes(self, *entity_classes):
        """Takes the given list of Entity classes, caches each one plus its
        associated model and also builds a mapping where the keys are GraphQL
        type names and the values are dicts containing the Entity and Model classes.

        Args:
            entity_classes: list of Entity classes for which we want to build
                Graphene types.
        """
        # Entity classes for API settings
        for entity_class in entity_classes:
            model_class = entity_class.Meta.model

            self._model_cache[model_class] = {
                self._ENTITY_CLASS_KEY: entity_class,
            }

            # Update our mapping of names (of GraphQL schema objects) to
            # model class and entity class. During query optimization,
            # as we traverse the tree of selected fields in the GraphQL query,
            # we only have access to the GraphQL type names. This mapping
            # will enable us to map those names to the associated model.
            node_name = self.schema.get_node_name_for_model(model_class)
            self._node_graphql_type_name_to_model_class[node_name] = {
                "model_class": model_class,
                "entity_class": entity_class,
            }

    def get_or_create_type(self, type_key, **kwargs):
        """Given a type_key (from the Types class) and one or more kwargs,
        get or build the type corresponding to the type_key.

        The kwargs are passed through to the builder in the case we haven't yet
        built this type which is indicated by the type not being in the cache.

        Args:
            type_key: string from the djraphql.schemas.default.mappings.Types class.
        Kwargs:
            Arbitrary, but most of the time it will be "model_class". The kwargs
                are defined by the builder's make method. E.g. in basic.py, the
                model_class is required.
        Returns:
            Graphene type associated with the type_key
        """
        if type_key not in self.schema.BUILDER_MAP:
            raise Exception(
                "get_or_create_type called with unknown key: {}".format(type_key)
            )

        # Check to see if we've already built this type and if so, return it.
        builder = self.schema.BUILDER_MAP[type_key]
        cache_key = builder.cache_key(type_key, **kwargs)
        if cache_key in self._type_cache:
            return self._type_cache[cache_key]

        # The type was not in the cache, so use the builder class
        # for this type key to create it, then cache the result and return it.
        result = builder.make(self, type_key=type_key, **kwargs)
        self._type_cache[cache_key] = result
        return result

    def get_model_fields(self, model_class):
        """When building a type, a type builder class (e.g. basic.py) needs to
        know what fields the library user wants to expose on the model being
        processed. This method basically proxies to the Model class's associated
        Entity class which is the source of truth for what fields to expose.

        Since this operation may require iterating through each field on a Model
        class, we cache the result as this method is called many times per model.

        Args:
            model_class: class of the Django Model being processed
        Returns:
            dict where the keys are strings whose value is the field, and the
                values are djraphql.fields.ModelField instances.
        """

        if self._model_cache[model_class].get(self._MODEL_FIELDS_KEY):
            return self._model_cache[model_class].get(self._MODEL_FIELDS_KEY)

        result = {}
        entity_class = self.get_entity_class(model_class)
        available_model_classes = self.get_available_model_classes()
        for model_field in entity_class.get_all_model_fields():
            django_model_field = model_field.get_django_model_field()
            if (
                not django_model_field.is_relation
                or django_model_field.related_model in available_model_classes
            ):
                result[model_field.name] = model_field

        self._model_cache[model_class][self._MODEL_FIELDS_KEY] = result
        return result

    def get_pk_fields_for_model(self, model_class):
        """Convenience method that wraps a utility method that gets the primary-key
        fields on a model.
        """
        return get_pk_fields_for_model(model_class, self.get_available_model_classes())

    def get_graphene_type_or_default(self, model_field):
        """Convenience method that returns the appropriate Graphene type for a
        ModelField.

        Args:
            model_field: ModelField instance that _may_ have a graphene_type specified.
        Returns:
            A graphene type e.g. Int, String. If the passed-in ModelField specifies
                a graphene_type (set by the user), use that. Otherwise the schema's
                type mapping is used .
        """
        # If the ModelField specifies its desired graphene_type, use it
        if model_field.graphene_type:
            return model_field.graphene_type

        # Otherwise, fall back to the type map used by the generation schema
        field = model_field.get_django_model_field()
        return self.schema.FIELD_TYPE_MAP[field.get_internal_type()]



from abc import ABCMeta, abstractmethod
from six.moves import reduce

CREATE = 1
READ = 2
UPDATE = 4
DELETE = 8


class PermissionFlag:
    """Representation of a permission, such as Create. Can be passed a list of
    permission checks, which the context must satisfy in order to be able to perform
    the operation associated with this permission.
    """

    def __init__(self, *permission_checks):
        self.permission_checks = permission_checks

    def __and__(self, value):
        return self.TYPE & value

    def __rand__(self, value):
        return self.TYPE & value

    def __or__(self, value):
        return self.TYPE | value

    def __ror__(self, value):
        return self.TYPE | value

    def has_permission(self, context):
        for permission_check in self.permission_checks:
            if not permission_check.is_authorized(context):
                return False

        return True

    def matches(self, permission_type):
        return bool(self.TYPE & permission_type)


class AccessPermissionsBitmap:
    """This class is used to wrap a list of PermissionFlags, and allow bit-wise
    operations on the permissions passed in, which will be one of the 4 PermissionFlag
    objects defined below: Create, Read, Update, Delete.
    """

    def __init__(self, *permissions):
        self.permissions = permissions
        self._bitmap = reduce(lambda a, b: a | b, permissions)

    def __iter__(self):
        return iter(self.permissions)

    def __and__(self, value):
        return self._bitmap & value

    def __or__(self, value):
        return self._bitmap | value


class Create(PermissionFlag):
    TYPE = CREATE


class Read(PermissionFlag):
    TYPE = READ


class Update(PermissionFlag):
    TYPE = UPDATE


class Delete(PermissionFlag):
    TYPE = DELETE


"""Expose convenient constants that contain no checks. I.e., Permissions(C, R, U, D)
means the Entity will be C/R/U/D-able without restriction.
"""
C = Create()
R = Read()
U = Update()
D = Delete()


class RequestAuthorizationCheck:
    """Base class used to perform a context-level (i.e. the request object) check
    against a permission type, e.g. Create or Read.

    E.g. if Foos are readable by everyone, but only staff users can create them,
    then you'd need a class to check whether a user is staff:

    class IsStaffUser(RequestAuthorizationCheck)
        def is_authorized(self, context):
            return context.user.is_staff

    And define the permissions like:

    class FooEntity(Entity):
        access_permissions = Permissions(Create(IsStaffUser()), R)
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def is_authorized(self, context):
        pass
